from whad.esb.stack.llm.exceptions import LinkLayerTimeoutException
from whad.esb.stack.llm.constants import ESBRole
from whad.scapy.layers.esb import ESB_Hdr, ESB_Payload_Hdr, ESB_Ack_Response
from whad.common.stack import Layer, alias, source, state, LayerState, instance
from queue import Queue, Empty
from time import sleep, time

import logging
logger = logging.getLogger(__name__)

class LinkLayerState(LayerState):

    def __init__(self):
        super().__init__()

        self.pid = 0
        self.role = ESBRole.PTX

        self.synchronized = False
        self.promiscuous = False

        self.last_timestamp = time()

        self.ack_miss = 0


@alias('ll')
@state(LinkLayerState)
class LinkLayer(Layer):
    def configure(self, options={}):
        self.__ack_queue = Queue()
        self.__data_queue = Queue()

    @property
    def channel(self):
        return self.get_layer('phy').channel

    @channel.setter
    def channel(self, channel):
        self.get_layer('phy').channel = channel

    @property
    def address(self):
        return self.get_layer('phy').address

    @address.setter
    def address(self, address):
        self.get_layer('phy').address = address

    @property
    def app(self):
        return self.get_layer('app')

    @property
    def promiscuous(self):
        return self.state.promiscuous

    @promiscuous.setter
    def promiscuous(self, promiscuous):
        self.state.promiscuous = promiscuous

    @property
    def synchronized(self):
        return self.state.synchronized

    @property
    def role(self):
        return self.state.role

    @role.setter
    def role(self, role):
        self.state.role = role

    def _increment_pid(self):
        self.state.pid = (self.state.pid + 1) % 4

    def synchronize(self, timeout: float = 10.0):
        self.state.role = ESBRole.PTX
        self.channel = None
        start_time = time()
        self.__ack_queue.queue.clear()
        self.__data_queue.queue.clear()
        self.state.promiscuous = True
        queue = self.__ack_queue
        while (time() - start_time) < timeout:
            try:
                queue = self.__ack_queue if queue == self.__data_queue else self.__data_queue
                msg = queue.get(block=False,timeout=0.05)

                if hasattr(msg, "metadata") and hasattr(msg.metadata, "channel"):
                    self.channel = msg.metadata.channel

                    if not self.state.synchronized:
                        self.state.synchronized = True
                        self.on_synchronized()
                    break
            except Empty:
                pass
        self.state.promiscuous = False
        return self.state.synchronized

    @source('app', 'data')
    def send_data(self, data: bytes, waiting_ack: bool = False):
        self.state.role = ESBRole.PTX
        packet = ESB_Hdr(
                pid=self.state.pid,
                address=self.get_layer('phy').address,
                no_ack=0
        ) / ESB_Payload_Hdr() / data

        self.send('phy', packet, channel=self.get_layer('phy').channel)
        self._increment_pid()
        if waiting_ack:
            try:
                ack = self.wait_for_ack()
                self.state.ack_miss = 0
                self.state.synchronized = True
                return ack
            except LinkLayerTimeoutException:
                self.state.ack_miss += 1
                if self.state.ack_miss > 10:
                    self.state.ack_miss = 0
                    if self.state.synchronized:
                        self.state.synchronized = False
                        self.on_desynchronized()
                return None

    def wait_for_ack(self, timeout=0.1):
        self.state.role = ESBRole.PTX
        start_time = time()
        while (time() - start_time) < timeout:
            try:
                msg = self.__ack_queue.get(block=False,timeout=0.001)
                return msg
            except Empty:
                pass
        raise LinkLayerTimeoutException

    def wait_for_data(self, timeout=0.1):
        self.state.role = ESBRole.PRX
        start_time = time()
        while (time() - start_time) < timeout:
            try:
                msg = self.__data_queue.get(block=False,timeout=0.001)
                return msg
            except Empty:
                pass
        raise LinkLayerTimeoutException

    def data_stream(self):
        self.state.role = ESBRole.PRX
        while True:
            yield self.__data_queue.get(block=True)

    @source('app', 'ack')
    def prepare_acknowledgment(self, data):
        self.state.role = ESBRole.PRX
        packet = ESB_Hdr(
                pid=self.state.pid,
                address=self.get_layer('phy').address,
                no_ack=True
        ) / ESB_Payload_Hdr() / data
        self.send('phy', packet)

    def on_synchronized(self):
        if self.get_layer('app') is not None:
            self.send('app', time(), tag='synchronized')

    def on_desynchronized(self):
        if self.get_layer('app') is not None:
            self.send('app', time(), tag='desynchronized')


    def on_prx_pdu(self, pdu):
        if (self.state.role == ESBRole.PTX or self.state.promiscuous):
            self.__ack_queue.put(pdu)

            if self.get_layer('app') is not None and len(bytes(pdu)) > 0:
                self.send('app', pdu[ESB_Payload_Hdr:], tag='ack')

    def on_ptx_pdu(self, pdu):
        if (self.state.role == ESBRole.PRX or self.state.promiscuous):

            self.__data_queue.put(pdu)

            if self.get_layer('app') is not None:
                self.send('app', pdu[ESB_Payload_Hdr:], tag='data')

    @source('phy')
    def on_pdu(self, pdu):
        if ESB_Ack_Response in pdu or len(bytes(pdu)) == 0:
            self.on_prx_pdu(pdu)
        else:
            self.on_ptx_pdu(pdu)



"""Bluetooth Low Energy Device Information Service Profile
"""
from struct import pack, unpack
from whad.ble.profile.attribute import UUID
from whad.ble.profile import PrimaryService, Characteristic

class DeviceInformationService(object):
    """Define Device Information Service
    """

    device_info = PrimaryService(
        uuid=UUID(0x180a),
        manufacturer=Characteristic(
            uuid=UUID(0x2A29),
            permissions=['read'],
        ),
        model_number=Characteristic(
            uuid=UUID(0x2A24),
            permissions=['read']
        ),
        serial_number=Characteristic(
            uuid=UUID(0x2A25),
            permissions=['read']
        ),
        hw_revision=Characteristic(
            uuid=UUID(0x2A27),
            permissions=['read']
        ),
        fw_revision=Characteristic(
            uuid=UUID(0x2A26),
            permissions=['read']
        ),
        sw_revision=Characteristic(
            uuid=UUID(0x2A28),
            permissions=['read']
        ),
        system_id=Characteristic(
            uuid=UUID(0x2A23),
            permissions=['read']
        ),
        ieee_rcdl=Characteristic(
            uuid=UUID(0x2A2A),
            permissions=['read']
        ),
        pnp_id=Characteristic(
            uuid=UUID(0x2A50),
            permissions=['read']
        )
    )

    @property
    def manufacturer(self):
        """Read device manufacturer
        """
        return self.device_info.manufacturer.value

    @manufacturer.setter
    def manufacturer(self, manufacturer: bytes):
        """Set device manufacturer name
        """
        self.device_info.manufacturer_name.value = manufacturer

    @property
    def model(self):
        """Read model number
        """
        return self.device_info.model_number.value

    @model.setter
    def model(self, model_number: bytes):
        self.device_info.model_number.value = model_number

    @property
    def serial(self):
        """Read device serial number
        """
        return self.device_info.serial_number.value

    @serial.setter
    def serial(self, serial_number: bytes):
        """Set device serial number
        """
        self.device_info.serial_number.value = serial_number

    @property
    def hw_revision(self):
        """Return current hw revision string
        """
        return self.device_info.hw_revision.value

    @hw_revision.setter
    def hw_revision(self, hw_revision_value: bytes):
        """Set device HW revision string
        """
        self.device_info.hw_revision.value = hw_revision_value

    @property
    def fw_revision(self):
        """Return current fw_revision string
        """
        return self.device_info.fw_revision.value

    @fw_revision.setter
    def fw_revision(self, fw_revision_value: bytes):
        """Set fw revision
        """
        self.device_info.fw_revision.value = fw_revision_value

    @property
    def sw_revision(self):
        """Return current sw_revision string
        """
        return self.device_info.sw_revision.value

    @sw_revision.setter
    def sw_revision(self, sw_revision_value: bytes):
        """Set fw revision
        """
        self.device_info.sw_revision.value = sw_revision_value

    @property
    def system_id(self):
        """Return system id
        """
        return self.device_info.system_id.value

    @system_id.setter
    def system_id(self, sysid_value: bytes):
        """Set system ID
        """
        self.device_info.system_id.value = sysid_value


"""
BLE Scanner connector
=====================

This module provides a scanner connector :class:`whad.ble.connector.scanner.Scanner`
for WHAD BLE devices, allowing to discover available BLE devices.

It implements two different scanning approaches:

* The first one is based on a *normal device scanning* based on WHAD BLE protocol:
  it puts the hardware adapter in device discovery mode and grab each advertisement
  received.

* The second one puts the hardware adapter in *sniffing mode* and sniffs advertisements
  sent on default channels. This mode is only available with WHAD devices that support
  sniffing.

This connector must be instantiated with a compatible WHAD device, as shown below:

.. code-block:: python

    device = WhadDevice.create('uart0')
    scanner = Scanner(device)

If the underlying device does not support scanning, this connector will raise
an :class:`UnsupportedCapability` exception.

"""
from scapy.packet import Packet
from typing import Iterator
from whad.hub.ble import BleAdvPduReceived, BleRawPduReceived
from whad.ble.connector import BLE
from whad.ble.scanning import AdvertisingDevicesDB, AdvertisingDevice
from whad.ble import UnsupportedCapability, message_filter, BleAdvType,\
    BTLE_ADV_IND, BTLE_ADV_DIRECT_IND, BTLE_ADV_NONCONN_IND, BTLE_ADV_SCAN_IND,\
    BTLE_SCAN_RSP, BTLE_ADV

class Scanner(BLE):
    """
    BLE Observer interface for compatible WHAD device.
    """

    domain = 'ble'

    def __init__(self, device):
        """Instantiate scanner connector over `device`.

        :param  device: BLE WHAD device instance
        :type   device: :class:`whad.device.WhadDevice`
        """
        super().__init__(device)
        self.__db = AdvertisingDevicesDB()

        # Check device accept scanning mode
        if not self.can_scan():
            # Does our device accept sniffing mode and is able of sniffing advertising reports ?
            if not self.can_sniff_advertisements():
                raise UnsupportedCapability('Scan')
            self.stop()
            self.sniff_advertisements()
        else:
            self.stop()
            self.enable_scan_mode(True)

    def start(self):
        """Start the BLE scanner.

        Calling this method resets the discovered devices database and put
        the WHAD device into BLE scanning mode.
        """
        self.__db.reset()
        super().start()


    def discover_devices(self, minimal_rssi = None, filter_address = None) -> Iterator[AdvertisingDevice]:
        """
        Parse incoming advertisements and yield discovered devices.

        :param  minimal_rssi:       Minimal RSSI level
        :type   minimal_rssi:       float, optional
        :param  filter_address:     BD address of a device to discover
        :type   filter_address:     :class:`whad.ble.bdaddr.BDAddress`, optional
        """
        for advertisement in self.sniff():
            if minimal_rssi is None or advertisement.metadata.rssi > minimal_rssi:
                devices = self.__db.on_device_found(
                    advertisement.metadata.rssi,
                    advertisement,
                    filter_addr=filter_address
                )
                for device in devices:
                    if device is not None and device.scanned:
                        yield device


    def sniff(self) -> Iterator[Packet]:
        """
        Listen and yield incoming advertising PDUs.
        """

        while True:
            if self.support_raw_pdu():
                message_type = BleRawPduReceived
            else:
                message_type = BleAdvPduReceived

            message = self.wait_for_message(filter=message_filter(message_type))
            # Convert message from rebuilt PDU
            packet = message.to_packet()
            self.monitor_packet_rx(packet)
            # Force TxAdd value to propagate the address type
            if isinstance(message, BleAdvPduReceived):
                if message.addr_type > 0:
                    packet.getlayer(BTLE_ADV).TxAdd = 1
            yield packet

    def clear(self):
        """
        Clear device database.
        """
        self.__db.reset()



"""This module provides a VirtuaDevice class that can be used with a WhadDeviceConnector
to interact with a device that doesn't support WHAD protocol. It allows to convert WHAD messages
to the corresponding specific API calls.

This class handles device connection, disconnection and read/write operations. All the
parsing magic is performed in our WhadDevice class.
"""

from asyncio import QueueEmpty
import os
import select
from threading import Lock
from time import sleep
from queue import Empty

from whad.device import WhadDevice
from whad.exceptions import WhadDeviceNotReady
from whad.protocol.whad_pb2 import Message
from whad.helpers import message_filter,is_message_type
from whad.protocol.device_pb2 import DeviceResetQuery,DeviceType
from whad.protocol.generic_pb2 import ResultCode
from whad.exceptions import WhadDeviceNotFound

import logging
logger = logging.getLogger(__name__)

class VirtualDevice(WhadDevice):
    """
    AdapterDevice device class.
    """
    def __init__(self):
        self._dev_type = None
        self._dev_id = None
        self._fw_author = None
        self._fw_url = None
        self._fw_version = None
        self._dev_capabilities = None
        self.__lock = Lock()
        super().__init__()

    def send_message(self, message, keep=None):
        self.__lock.acquire()
        super().set_queue_filter(keep)
        self._on_whad_message(message)
        self.__lock.release()

    def _on_whad_message(self, message):
        """TODO: associate callbacks with classes ?
        """
        category = message.message_type
        message_type = message.message_name

        callback_name = "_on_whad_"+category+"_"+message_type
        if hasattr(self, callback_name) and callable(getattr(self, callback_name)):
            getattr(self, callback_name)(message)
        else:
            logger.info("unhandled message: %s" % message)
            self._send_whad_command_result(ResultCode.ERROR)

    def _on_whad_discovery_info_query(self, message):
        major, minor, revision = self._fw_version
        msg = self.hub.discovery.create_info_resp(
            DeviceType.VirtualDevice,
            self._dev_id,
            0x0100,
            0,
            self._fw_author,
            self._fw_url,
            major, minor, revision,
            [domain | (capabilities[0] & 0xFFFFFF) for domain, capabilities in self._dev_capabilities.items()]
        )
        self._send_whad_message(msg)

    def _on_whad_discovery_domain_query(self, message):
        # Compute supported commands for domain
        commands = 0
        supported_commands = self._dev_capabilities[message.domain][1]
        for command in supported_commands:
            commands |= (1 << command)

        # Create a DomainResp message and send it
        msg = self.hub.discovery.create_domain_resp(
            message.domain,
            commands
        )
        self._send_whad_message(msg)


    def _send_whad_message(self, message):
        self.on_message_received(message)

    def _send_whad_command_result(self, code):
        msg = self.hub.generic.create_command_result(code)
        self._send_whad_message(msg)

from .ubertooth import UbertoothDevice
from .rzusbstick import RZUSBStickDevice
from .apimote import APIMoteDevice
from .hci import HCIDevice
from .rfstorm import RFStormDevice
from .yard import YardStickOneDevice
from .pcap import PCAPDevice



from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady, WhadDeviceAccessDenied, \
    WhadDeviceDisconnected
from whad.device.virtual import VirtualDevice
from whad.helpers import message_filter,is_message_type,bd_addr_to_bytes
from whad.device.virtual.pcap.capabilities import CAPABILITIES
from whad.hub.generic.cmdresult import CommandResult
from whad.hub.dot15d4 import Commands
from scapy.layers.dot15d4 import Dot15d4#, Dot15d4FCS
from whad.ble.utils.phy import channel_to_frequency, frequency_to_channel, crc, FieldsSize, is_access_address_valid
from scapy.utils import PcapReader, PcapWriter
from struct import unpack, pack
from scapy.layers.bluetooth4LE import BTLE
from whad.scapy.layers.phy import Phy_Packet
from whad.hub.dot15d4 import Dot15d4Metadata
from whad.hub.ble import BLEMetadata
from whad.hub.esb import ESBMetadata
from whad.hub.phy import PhyMetadata, Modulation, Endianness
from whad.hub.unifying import UnifyingMetadata
from time import sleep
from whad.hub.discovery import Domain
from os.path import exists
import logging
import sys

logger = logging.getLogger(__name__)

class PCAPDevice(VirtualDevice):

    INTERFACE_NAME = "pcap"

    @classmethod
    def list(cls):
        '''
        Returns a list of available PCAP devices.
        '''
        return None

    @classmethod
    def check_interface(cls, interface):
        '''
        This method checks dynamically if the provided interface can be instantiated.
        '''
        logger.info("Checking interface: %s" % str(interface))
        if interface.endswith(".pcap") or interface.endswith(".pcapng"):
            return True
        else:
            return False

    @property
    def identifier(self):
        '''
        Returns the identifier of the current device (e.g., bus + address in format "<bus>-<address>").
        '''
        return "pcap:" + str(self.__filename)


    def __init__(self, filename):
        """
        Create device connection
        """
        self.__opened = False
        self.__started = False
        self.__flush = "flush:" in filename
        self.__filename = filename.replace("flush:", "")
        self.__pcap_reader = None
        self.__pcap_writer = None
        self.__dlt = None
        self.__domain = None
        self.__start_timestamp, self.__last_timestamp = None, None


        self.__supported_frequency_range = [
            (281000000, 361000000),
            (378000000, 481000000),
            (749000000, 962000000),
            (2400000000, 2500000000)
        ]
        super().__init__()

    def _is_reader(self):
        """
        Returns True if the PCAP is in reading mode.
        """
        return self.__pcap_reader is not None

    def _get_dlt(self):
        if self._is_reader():
            if hasattr(self.__pcap_reader, "linktype"):
                dlt = self.__pcap_reader.linktype
            else:
                # PCAP-ng
                _ = self.__pcap_reader.read_packet()
                dlt = self.__pcap_reader.interfaces[0][0]
                self.__pcap_reader = PcapReader(self.__filename)
            return dlt

    def _get_domain(self):
        return list(CAPABILITIES[self.__dlt][0].keys())[0]

    def open(self):
        try:
            #print("Opening:", self.__filename)
            if exists(self.__filename):
                logger.info("Existing PCAP file")
                self.__pcap_reader = PcapReader(self.__filename)
                self.__dlt = self._get_dlt()
                self.__domain = self._get_domain()
            else:
                logger.info("No PCAP file")
                raise WhadDeviceNotFound("pcap")

        except:
            raise WhadDeviceAccessDenied("pcap")

        self._dev_id = self._get_serial_number()
        self._fw_author = self._get_manufacturer()
        self._fw_url = self._get_url()
        self._fw_version = self._get_firmware_version()
        self._dev_capabilities = self._get_capabilities()
        #self.__flush = False
        self.__opened = True
        # Ask parent class to run a background I/O thread
        super().open()

    def write(self, data):
        if not self.__opened:
            raise WhadDeviceNotReady()

    def read(self):
        if not self.__opened:
            raise WhadDeviceNotReady()
        while self.__started:
            try:
                if self._is_reader():
                    pkt = self.__pcap_reader.read_packet()
                    self._send_packet(pkt)
            except EOFError:
                # TODO: add an event to indicate end of stream ?
                logger.debug('[PCAPDevice] EOF reached')
                raise WhadDeviceDisconnected()

    def reset(self):
        pass

    def _generate_metadata(self, pkt):
        if self.__domain == Domain.Dot15d4:
            metadata = Dot15d4Metadata.convert_from_header(pkt)
        elif self.__domain == Domain.BtLE:
            metadata = BLEMetadata.convert_from_header(pkt)
        elif self.__domain == Domain.Esb:
            metadata = ESBMetadata.convert_from_header(pkt)
        elif self.__domain == Domain.LogitechUnifying:
            metadata = UnifyingMetadata.convert_from_header(pkt)
        elif self.__domain == Domain.Phy:
            metadata = PhyMetadata.convert_from_header(pkt)

        else:
            return None
        if self.__start_timestamp is None:
            self.__start_timestamp = metadata.timestamp
        metadata.timestamp = metadata.timestamp - self.__start_timestamp
        return metadata

    def _interframe_delay(self, timestamp):
        if not self.__flush:
            if self.__last_timestamp is None:
                self.__last_timestamp = 0
            sleep((timestamp - self.__last_timestamp)/100000)

    def _send_packet(self, pkt):
        if self.__domain == Domain.Dot15d4:
            metadata = self._generate_metadata(pkt)
            self._interframe_delay(metadata.timestamp)
            self.__last_timestamp = metadata.timestamp
            self._send_whad_zigbee_raw_pdu(bytes(pkt[Dot15d4]), channel=metadata.channel, lqi=metadata.lqi, rssi=metadata.rssi, timestamp=metadata.timestamp)
        elif self.__domain == Domain.BtLE:
            metadata = self._generate_metadata(pkt)
            self._interframe_delay(metadata.timestamp)
            self.__last_timestamp = metadata.timestamp
            self._send_whad_ble_raw_pdu(pkt, metadata)
        elif self.__domain == Domain.Esb:
            metadata = self._generate_metadata(pkt)
            self._interframe_delay(metadata.timestamp)
            self.__last_timestamp = metadata.timestamp
            self._send_whad_esb_raw_pdu(pkt, metadata)
        elif self.__domain == Domain.LogitechUnifying:
            metadata = self._generate_metadata(pkt)
            self._interframe_delay(metadata.timestamp)
            self.__last_timestamp = metadata.timestamp
            self._send_whad_unifying_raw_pdu(pkt, metadata)
        elif self.__domain == Domain.Phy:
            metadata = self._generate_metadata(pkt)
            self._interframe_delay(metadata.timestamp)
            self.__last_timestamp = metadata.timestamp
            self._send_whad_phy_pdu(pkt, metadata)


    def _send_whad_phy_pdu(self, packet, metadata):
        #packet.show()
        msg = self.hub.phy.create_packet_received(
            metadata.frequency, # TODO: frequency,
            bytes(packet[Phy_Packet]),
            metadata.rssi, # TODO: rssi
            metadata.timestamp,
            metadata.syncword,
            metadata.datarate,
            metadata.deviation,
            Modulation(metadata.modulation),
            Endianness(metadata.endianness)
        )
        # Send message
        self._send_whad_message(msg)

    def _send_whad_unifying_raw_pdu(self, packet, metadata):
        # Create a RawPduReceived message
        msg = self.hub.unifying.create_raw_pdu_received(
            metadata.channel,
            bytes(packet),
            metadata.rssi,
            metadata.timestamp,
            metadata.is_crc_valid,
            metadata.address
        )

        # Send message
        self._send_whad_message(msg)

    def _send_whad_esb_raw_pdu(self, packet, metadata):

        # Create a RawPduReceived message
        msg = self.hub.esb.create_raw_pdu_received(
            metadata.channel,
            bytes(packet),
            metadata.rssi,
            metadata.timestamp,
            metadata.is_crc_valid,
            metadata.address
        )

        # Send message
        self._send_whad_message(msg)

    def _send_whad_ble_raw_pdu(self, packet, metadata):
        packet = packet[BTLE:]
        access_address = packet.access_addr
        pdu = bytes(packet)[FieldsSize.ACCESS_ADDRESS_SIZE:-FieldsSize.CRC_SIZE]

        # Create a RawPduReceived message
        msg = self.hub.ble.create_raw_pdu_received(
            metadata.direction,
            pdu,
            access_address,
            0,
            crc_validity=metadata.is_crc_valid,
            crc=packet.crc,
            channel=metadata.channel,
            timestamp=metadata.timestamp,
            rssi=metadata.rssi
        )

        # Send message
        self._send_whad_message(msg)



    # Virtual device whad message builder
    def _send_whad_zigbee_raw_pdu(self, packet, channel=None, rssi=None, lqi=None, is_fcs_valid=True, timestamp=None):
        pdu = packet[:-2]
        fcs = unpack("<H",packet[-2:])[0]

        # Create a RawPduReceived message
        msg = self.hub.dot15d4.create_raw_pdu_received(
            channel,
            pdu,
            fcs,
            lqi = lqi,
            fcs_validity=is_fcs_valid
        )

        # Set timestamp and RSSI if provided
        if rssi is not None:
            msg.rssi = rssi
        if timestamp is not None:
            msg.timestamp = timestamp

        # Send message
        self._send_whad_message(msg)


    # Virtual device whad message callbacks
    def _on_whad_ble_stop(self, message):
        self.__started = False
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_ble_start(self, message):
        self.__started = True
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_ble_sniff_adv(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_ble_sniff_conn(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)


    def _on_whad_phy_stop(self, message):
        self.__started = False
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_send_raw(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_sniff(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_start(self, message):
        self.__started = True
        self._send_whad_command_result(CommandResult.SUCCESS)


    def _on_whad_phy_get_supported_freq(self, message):
        # Create a SupportedFreqRanges
        msg = self.hub.phy.create_supported_freq_ranges(
            self.__supported_frequency_range
        )

        # Send message
        self._send_whad_message(msg)

    def _on_whad_phy_send(self, message):
        self._send_packet(message.packet)
        self._send_whad_command_result(CommandResult.SUCCESS)


    def _on_whad_phy_set_freq(self, message):
        found = False
        for supported_range in self.__supported_frequency_range:
            range_start, range_end = supported_range[0], supported_range[1]
            if message.frequency >= range_start and message.frequency <= range_end:
                found = True
                break
        if found:
            self.__frequency = message.frequency
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.PARAMETER_ERROR)

    def _on_whad_phy_datarate(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_packet_size(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_endianness(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_tx_power(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_mod_ask(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_mod_4fsk(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_mod_fsk(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_mod_gfsk(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_sync_word(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)


    def _on_whad_dot15d4_stop(self, message):
        self.__started = False
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_dot15d4_send_raw(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_dot15d4_sniff(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_dot15d4_start(self, message):
        self.__started = True
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_esb_stop(self, message):
        self.__started = False
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_esb_send_raw(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_esb_sniff(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_esb_start(self, message):
        self.__started = True
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_unifying_stop(self, message):
        self.__domain = Domain.LogitechUnifying
        self.__started = False
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_unifying_send_raw(self, message):
        self.__domain = Domain.LogitechUnifying
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_unifying_sniff(self, message):
        self.__domain = Domain.LogitechUnifying
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_unifying_start(self, message):
        self.__domain = Domain.LogitechUnifying
        self.__started = True
        self._send_whad_command_result(CommandResult.SUCCESS)

    # Discovery related functions
    def _get_capabilities(self):
        index = 0 if self._is_reader() else 1
        capabilities = CAPABILITIES[self.__dlt][index]
        return capabilities

    def _get_manufacturer(self):
        return "whad-team".encode('utf-8')

    def _get_serial_number(self):
        return bytes.fromhex("00" * 16)

    def _get_firmware_version(self):
        return (0, 0, 0)

    def _get_url(self):
        return "https://github.com/whad-team/whad-client".encode('utf-8')



"""WHAD server tool

This utility implements a server module, allowing to create a TCP proxy
which can be used to access a device remotely.
"""
from prompt_toolkit import print_formatted_text, HTML

from whad.common.monitors.pcap import PcapWriterMonitor
from whad.cli.app import CommandLineApp, ApplicationError, run_app
from scapy.all import *
from whad.device.unix import UnixSocketServerDevice, UnixConnector
from whad.device import Bridge, ProtocolHub
from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady
from whad.cli.ui import error, warning, success, info, display_event, display_packet
import logging
from time import sleep
from scapy.config import conf

logger = logging.getLogger(__name__)
#logging.basicConfig(level=logging.DEBUG)

class WhadFilterPipe(Bridge):
    """Whad filter output pipe

    When wfilter is chained with another whad tool, it spawns a device
    based on the specified profile using the specified WHAD adapter and forward
    it to the chained tool. The chained tool will then forward packets back and forth.
    """
    def __init__(self, input_connector, output_connector, on_rx_packet_cb, on_tx_packet_cb):
        super().__init__(input_connector, output_connector)
        self.on_rx_packet = on_rx_packet_cb
        self.on_tx_packet = on_tx_packet_cb


    def on_outbound(self, message):
        """Process outbound messages.

        Outbound packets are packets coming from our input connector,that need to be
        forwarded as packets to the next tool.
        """
        if hasattr(message, "to_packet"):
            pkt = message.to_packet()
            pkt = self.on_rx_packet(pkt)
            if pkt is not None:
                msg = message.from_packet(pkt)
                super().on_outbound(msg)
        else:
            logger.debug('[wfilter][input-pipe] forward default outbound message %s' % message)
            # Forward other messages
            super().on_outbound(message)


    def on_inbound(self, message):
        """Process inbound messages.

        Inbound packets are packets coming from our output connector,that need to be
        forwarded as packets to the previous tool.
        """
        if hasattr(message, "to_packet"):
            pkt = message.to_packet()
            pkt = self.on_tx_packet(pkt)
            if pkt is not None:
                msg = message.from_packet(pkt)
                super().on_inbound(msg)
        else:
            logger.debug('[wfilter][input-pipe] forward default inbound message %s' % message)
            # Forward other messages
            super().on_inbound(message)


class WhadFilterApp(CommandLineApp):

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD filter tool',
            interface=True,
            commands=False,
            input=CommandLineApp.INPUT_WHAD,
            output=CommandLineApp.OUTPUT_WHAD
        )

        self.add_argument(
            'filter',
            help='filter to evaluate',
            nargs='?'
        )

        self.add_argument(
            '--down',
            dest='down',
            action="store_true",
            default=None,
            help='process down stream'
        )


        self.add_argument(
            '--up',
            dest='up',
            action="store_true",
            default=None,
            help='process up stream'
        )

        self.add_argument(
            '-t',
            '--transform',
            dest='transform',
            type=str,
            default=None,
            help='apply a transformation'
        )

        self.add_argument(
            '-e',
            '--invert',
            dest='invert',
            action="store_true",
            default=False,
            help='invert filter'
        )

        self.add_argument(
            '-f',
            '--forward',
            dest='forward',
            action="store_true",
            default=False,
            help='forward packets not matched by the filter (dropped by default)'
        )
    def build_filter(self):
        if self.args.filter is None:
            self.args.filter = "True"
        filter_template = "lambda p : not %s" if self.args.invert else "lambda p : %s"
        if "packet." in self.args.filter:
            self.args.filter.replace("packet.", "p.")
        elif "pkt." in self.args.filter:
            self.args.filter.replace("pkt.", "p.")

        return eval(filter_template % (self.args.filter))

    def on_rx_packet(self, pkt):
        if not self.args.down:
            if not self.is_stdout_piped():
                display_packet(
                    pkt,
                    show_metadata = self.args.metadata,
                    format = self.args.format
                )
            return pkt

        filter = self.build_filter()
        try:
            if filter(pkt):
                if self.args.transform is not None:
                    p = packet = pkt
                    exec(self.args.transform)
                    # recompute CRC ?
                if not self.is_stdout_piped():
                    display_packet(
                        pkt,
                        show_metadata = self.args.metadata,
                        format = self.args.format
                    )
                return pkt
            else:

                if self.args.forward:
                    display_packet(
                        pkt,
                        show_metadata = self.args.metadata,
                        format = self.args.format
                    )
                    return pkt
                return None
        except:
            if self.args.forward:
                display_packet(
                    pkt,
                    show_metadata = self.args.metadata,
                    format = self.args.format
                )
                return pkt
            return None

    def on_tx_packet(self, pkt):
        if not self.args.up:
            if not self.is_stdout_piped():
                display_packet(
                    pkt,
                    show_metadata = self.args.metadata,
                    format = self.args.format
                )
            return pkt

        filter = self.build_filter()
        try:
            if filter(pkt):
                if self.args.transform is not None:
                    p = packet = pkt
                    exec(self.args.transform)
                if not self.is_stdout_piped():
                    display_packet(
                        pkt,
                        show_metadata = self.args.metadata,
                        format = self.args.format
                    )
                return pkt
            else:
                if self.args.forward:
                    display_packet(
                        pkt,
                        show_metadata = self.args.metadata,
                        format = self.args.format
                    )
                    return pkt
                return None
        except:
            if self.args.forward:
                display_packet(
                    pkt,
                    show_metadata = self.args.metadata,
                    format = self.args.format
                )
                return pkt
            return None

    def run(self):
        # Launch pre-run tasks
        self.pre_run()
        if self.args.down is None and self.args.up is None:
            self.args.up   = True
            self.args.down = True

        try:
            if self.is_piped_interface():
                interface = self.input_interface
            else:
                interface = self.interface

            if interface is not None:
                if not self.args.nocolor:
                    conf.color_theme = BrightTheme()

                parameters = self.args.__dict__
                connector = UnixConnector(interface)

                connector.domain = self.args.domain
                hub = ProtocolHub(2)
                connector.format = hub.get(self.args.domain).format

                #connector.translator = get_translator(self.args.domain)(connector.hub)
                #connector.format = connector.translator.format

                if self.is_stdout_piped():
                    unix_server = UnixConnector(UnixSocketServerDevice(parameters={
                        'domain': self.args.domain,
                        'format': self.args.format,
                        'metadata' : self.args.metadata
                    }))


                    while not unix_server.device.opened:
                        if unix_server.device.timedout:
                            return
                        else:
                            sleep(0.1)
                    # Create our packet bridge
                    logger.info("[wfilter] Starting our output pipe")
                    output_pipe = WhadFilterPipe(connector, unix_server, self.on_rx_packet, self.on_tx_packet)

                else:
                    connector.on_packet = self.on_rx_packet

                # Keep running while interface is active
                while interface.opened:
                    sleep(.1)
            else:
                exit(1)
        except KeyboardInterrupt:
            # Launch post-run tasks
            self.post_run()


def wfilter_main():
    app = WhadFilterApp()
    run_app(app)



"""Bluetooth Low Energy MitM tool

This utility will connect to a device and spawn a similar device that will act
as a proxy to the target device.
"""
from time import time
from hexdump import hexdump
from prompt_toolkit import print_formatted_text, HTML

from whad.ble import Scanner
from whad.device import WhadDevice
from whad.ble.tools.proxy import GattProxy, LinkLayerProxy
from whad.cli.app import CommandLineDeviceSource, run_app
from whad.hub.ble import Direction

import logging
logger = logging.getLogger(__name__)

class VerboseLLProxy(LinkLayerProxy):
    """Verbose link-layer proxy
    """

    def __init__(self, app, proxy=None, target=None, adv_data=None, scan_data=None, bd_address=None, spoof=False):
        """Initialize our parent class instance.
        """
        super().__init__(proxy=proxy, target=target, adv_data=adv_data, scan_data=scan_data, \
                         bd_address=bd_address, spoof=spoof)
        self.__app = app

    def on_connect(self):
        print_formatted_text(HTML(
            f"<ansimagenta>Remote device connected</ansimagenta>"
        ))

    def on_disconnect(self):
        print_formatted_text(HTML(
            f"<ansimagenta>Remote device disconnected</ansimagenta>"
        ))

    def on_ctl_pdu(self, pdu, direction):
        """Display captured Control PDU"""
        if direction == Direction.MASTER_TO_SLAVE:
            print_formatted_text(HTML("&lt;&lt;&lt; <ansicyan>Control PDU</ansicyan>"))      
        else:
            print_formatted_text(HTML("&gt;&gt;&gt; <ansicyan>Control PDU</ansicyan>"))
        hexdump(bytes(pdu))
        return super().on_ctl_pdu(pdu, direction)

    
    def on_data_pdu(self, pdu, direction):
        """Display captured data PDU"""
        if direction == Direction.MASTER_TO_SLAVE:
            print_formatted_text(HTML("&lt;&lt;&lt; <ansimagenta>Data PDU</ansimagenta>"))      
        else:
            print_formatted_text(HTML("&gt;&gt;&gt; <ansimagenta>Data PDU</ansimagenta>"))
        hexdump(bytes(pdu))
        return super().on_data_pdu(pdu, direction)

class VerboseProxy(GattProxy):

    def __init__(self, app, proxy=None, target=None, adv_data=None, scan_data=None, bd_address=None, spoof=False, profile=None):
        """Initialize our parent class instance.
        """
        super().__init__(proxy=proxy, target=target, adv_data=adv_data, scan_data=scan_data, \
                         bd_address=bd_address, spoof=spoof, profile=profile)
        self.__app = app
        
    def on_characteristic_read(self, service, characteristic, value, offset=0, length=0):
        if offset > 0:
            print_formatted_text(HTML(
                f"&lt;&lt;&lt; <ansicyan>Characteristic {characteristic.uuid} read (offset: {offset})</ansicyan>"
            ))
        else:
            print_formatted_text(HTML(
                f"&gt;&gt;&gt; <ansicyan>Characteristic {characteristic.uuid} written</ansicyan>"
            ))            
        hexdump(value)

    def on_characteristic_write(self, service, characteristic, offset=0, value=b'', without_response=False):
        if offset > 0:
            print_formatted_text(HTML(
                f"&lt;&lt;&lt; <ansicyan>Characteristic {characteristic.uuid} read (offset: {offset})</ansicyan>"
            ))
        else:
            print_formatted_text(HTML(
                f"&gt;&gt;&gt; <ansicyan>Characteristic {characteristic.uuid} written</ansicyan>"
            ))  
        hexdump(value)

    def on_characteristic_subscribed(self, service, characteristic, notification=False, indication=False):
        if notification:
            print_formatted_text(HTML(
                    f"[!] <ansicyan>Subscribed to notification for charac. {characteristic.uuid}</ansicyan>"
            ))
        if indication:
            print_formatted_text(HTML(
                    f"[!] <ansicyan>Subscribed to notification for charac. {characteristic.uuid}</ansicyan>"
            ))    

    def on_characteristic_unsubscribed(self, service, characteristic):
        print_formatted_text(HTML(
                f"[!] <ansicyan>Unubscribed from charac. {characteristic.uuid}</ansicyan>"
        ))

    def on_notification(self, service, characteristic, value):
        print_formatted_text(HTML(
            f"&lt;&lt;&lt; <ansicyan>[!] Notification for charac. {characteristic.uuid}:</ansicyan>"
        ))
        hexdump(value)

    def on_indication(self, service, characteristic, value):
        print_formatted_text(HTML(
            f"&lt;&lt;&lt; <ansicyan>[!] Indication for charac. {characteristic.uuid}:</ansicyan>"
        ))
        hexdump(value)

    def on_connect(self, conn_handle):
        print_formatted_text(HTML(
            f"<ansimagenta>Remote device connected</ansimagenta>"
        ))

    def on_disconnect(self, conn_handle):
        print_formatted_text(HTML(
            f"<ansimagenta>Remote device disconnected</ansimagenta>"
        ))

class BleProxyApp(CommandLineDeviceSource):
    """Bluetooth Low-Energy Proxy application

    This application provides a GATT proxy feature that allows the user to connect
    to a target device, discovers its services and characteristics and spawns a
    duplicate device and proxify any BLE operation to the target device.
    """

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD Bluetooth Low Energy proxy tool',
            interface=True,
            commands=False
        )

        self.add_argument("bdaddr", metavar="BDADDR", help="Target device BD address")

        # Add an optional random type argument
        self.add_argument(
            "-p",
            "--proxy-interface",
            dest="proxy_iface",
            type=str,
            help="Specify the WHAD interface to use for our spoofed device"
        )

        self.add_argument(
            "-t",
            "--timeout",
            dest="timeout",
            type=int,
            default=30,
            help="Scan timeout in seconds"
        )

        self.add_argument(
            "-w",
            "--wireshark",
            dest="wireshark",
            action="store_true",
            default=False,
            help="Enable real-time wireshark monitoring"
        )

        self.add_argument(
            "-s",
            "--spoof",
            dest="spoof",
            action="store_true",
            default=False,
            help="Enable BD address spoofing (if available)"
        )

        self.add_argument(
            "--link-layer",
            dest="linklayer",
            action="store_true",
            default=False,
            help="Enable link-layer mode"
        )

        self.add_argument(
            "--output",
            dest="output",
            default=None,
            help="Output PCAP file path"
        )

    def run(self):
        """Override App's run() method to handle scripting feature.
        """
        try:
            # Launch pre-run tasks
            self.pre_run()

            # We need to have an interface specified
            if self.interface is not None:
                self.spawn_proxy()                
            else:
                self.error("You need to specify an interface with option --interface.")

        except KeyboardInterrupt as keybd:
            self.warning("wble-proxy stopped (CTL-C)")

        # Launch post-run tasks
        self.post_run()

    def spawn_proxy(self):
        """Create a GATT proxy
        """
        proxy_iface = WhadDevice.create(self.args.proxy_iface)
        adv_data = None
        scan_rsp = None
        
        # Start scanning, we are looking for our target device
        print(f"Scanning for target device (timeout: {self.args.timeout} seconds)...")
        scan_start_ts = time()
        scanner = Scanner(self.interface)
        scanner.start()
        for device in scanner.discover_devices():
            if device.address.lower() == self.args.bdaddr:
                if device.adv_records is not None and device.scan_rsp_records is not None:
                    adv_data, scan_rsp = device.adv_records.to_bytes(), device.scan_rsp_records.to_bytes()
                    break
                
            # Device search timeout reached, show warning and stop proxy
            if time() - scan_start_ts > self.args.timeout:
                self.warning("Target device not found, connection timeout exceeded.")
                return

        if adv_data is not None and scan_rsp is not None:
            if not self.args.linklayer:
                proxy = VerboseProxy(
                    self,
                    proxy_iface,
                    self.interface,
                    adv_data=adv_data,
                    scan_data=scan_rsp,
                    bd_address=self.args.bdaddr,
                    spoof=self.args.spoof
                )
            else:
                proxy = VerboseLLProxy(
                    self,
                    proxy=proxy_iface,
                    target=self.interface,
                    adv_data=adv_data,
                    scan_data=scan_rsp,
                    bd_address=self.args.bdaddr,
                    spoof=self.args.spoof
                )
            
            # Start our proxy
            proxy.start()

            # Set output PCAP file if provided\
            if self.args.output is not None:
                print("Setting a pcap monitor")
                pcap_mon = proxy.get_pcap_monitor(self.args.output)
                pcap_mon.start()
            else:
                pcap_mon = None

            # Create a Wireshark monitor
            if self.args.wireshark:
                ws_mon = proxy.get_wireshark_monitor()
                ws_mon.start()
            else:
                ws_mon = None
            
            print('Proxy is ready, press a key to stop.')
            input()

            # Stop Wireshark monitor
            if ws_mon is not None:
                ws_mon.stop()
                ws_mon.detach()
            
            # Stop PCAP monitor
            if pcap_mon is not None:
                pcap_mon.stop()
                pcap_mon.detach()


def ble_proxy_main():
    app = BleProxyApp()
    run_app(app)



"""Att Exceptions
"""
class AttErrorCode:
    INVALID_HANDLE_VALUE = 0x01
    READ_NOT_PERMITTED = 0x02
    WRITE_NOT_PERMITTED = 0x03
    INVALID_PDU = 0x04
    INSUFFICIENT_AUTHENTICATION = 0x05
    UNSUPPORTED_REQUEST = 0x06
    INVALID_OFFSET = 0x07
    INSUFFICIENT_AUTHORIZATION = 0x08
    PREPARE_QUEUE_FULL = 0x09
    ATTR_NOT_FOUND = 0x0A
    ATTR_NOT_LONG = 0x0B
    INSUFFICIENT_ENC_KEY_SIZE = 0x0C
    INVALID_ATTR_VALUE_LENGTH = 0x0D
    UNLIKELY_ERROR = 0x0E
    INSUFFICIENT_ENC = 0x0F
    UNSUPPORTED_GROUP_TYPES = 0x10
    INSUFFICIENT_RES = 0x11

class AttError(Exception):
    """ATT generic exception
    """

    def __init__(self, request, handle):
        super().__init__()
        self.request = request
        self.handle = handle

class InvalidHandleValueError(AttError):
    """Exception raised when an invalid handle value is provided.
    """
    def __init__(self, request, handle):
        super().__init__(request, handle)

class ReadNotPermittedError(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

class WriteNotPermittedError(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

class InvalidPduError(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

class InsufficientAuthenticationError(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

class UnsupportedRequestError(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

class InvalidOffsetError(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

class InsufficientAuthorizationError(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

class PrepareQueueFullError(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

class AttributeNotFoundError(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

class AttributeNotLongError(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

class InsufficientEncryptionKeySize(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

class InvalidAttrValueLength(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

class UnlikelyError(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

class InsufficientEncryptionError(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

class UnsupportedGroupTypesError(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

class InsufficientResourcesError(AttError):
    def __init__(self, request, handle):
        super().__init__(request, handle)

def error_response_to_exc(error_code, request, handle):
    RESP_TO_EXC = {
        0x01: InvalidHandleValueError,
        0x02: ReadNotPermittedError,
        0x03: WriteNotPermittedError,
        0x04: InvalidPduError,
        0x05: InsufficientAuthenticationError,
        0x06: UnsupportedRequestError,
        0x07: InvalidOffsetError,
        0x08: InsufficientAuthorizationError,
        0x09: PrepareQueueFullError,
        0x0A: AttributeNotFoundError,
        0x0B: AttributeNotLongError,
        0x0C: InsufficientEncryptionKeySize,
        0x0D: InvalidAttrValueLength,
        0x0E: UnlikelyError,
        0x0F: InsufficientEncryptionError,
        0x10: UnsupportedGroupTypesError,
        0x11: InsufficientResourcesError
    }
    if error_code in RESP_TO_EXC:
        return RESP_TO_EXC[error_code](request, handle)
    else:
        return AttError(request, handle)



"""WHAD server tool

This utility implements a server module, allowing to create a TCP proxy
which can be used to access a device remotely.
"""
import logging
import time

from scapy.all import *
from scapy.config import conf
from whad.cli.ui import wait, success
from whad.device import Bridge, ProtocolHub
from whad.device.unix import UnixConnector, UnixSocketServerDevice
from whad.common.monitors import WiresharkMonitor
from whad.cli.app import CommandLineApp, run_app

# wshark logger
logger = logging.getLogger(__name__)


class WhadWiresharkApp(CommandLineApp):
    """Main WiresharkApp class
    """

    # App connector
    connector = None

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD wireshark',
            interface=True,
            commands=False,
            input=CommandLineApp.INPUT_WHAD,
            output=CommandLineApp.OUTPUT_WHAD
        )

    def run(self):
        """Application main() routine
        """

        # Launch pre-run tasks
        self.monitor = None
        self.pre_run()
        try:
            if self.is_piped_interface():
                interface = self.input_interface
            else:
                interface = self.interface

            if interface is not None:
                if not self.args.nocolor:
                    conf.color_theme = BrightTheme()

                parameters = self.args.__dict__

                parameters.update({
                    #"on_tx_packet_cb" : self.on_tx_packet,
                    #"on_rx_packet_cb" : self.on_rx_packet,
                })

                # Using a UnixConnector is just a small hack, as it acts like
                # a dummy connector.
                connector = UnixConnector(interface)
                if self.is_stdout_piped():
                    proxy = UnixConnector(UnixSocketServerDevice(parameters=self.args.__dict__))

                    while not proxy.device.opened:
                        if proxy.device.timedout:
                            return
                        else:
                            sleep(0.1)

                    # Bridge both connectors and their respective interfaces
                    bridge = Bridge(connector, proxy)

                # Save our connector and force its domain
                self.connector = connector
                self.connector.domain = self.args.domain
                #self.connector.translator = get_translator(self.args.domain)(connector.hub)
                #self.connector.format = connector.translator.format


                # Query our protocol hub to gather the correct format function
                # based on the provided domain
                hub = ProtocolHub(2)
                self.connector.format = hub.get(self.args.domain).format

                # Attack a wireshark monitor
                self.monitor = WiresharkMonitor()
                self.monitor.attach(self.connector)
                self.monitor.start()

                if self.is_stdout_piped():
                    # Wait for the user to CTL-C
                    while interface.opened:
                        time.sleep(.1)
                else:
                    while interface.opened:
                        wait("Forwarding {count} packets to wireshark".format(
                                count=str(self.monitor.packets_written)
                            )
                        )
                        time.sleep(.2)


        except KeyboardInterrupt:
            # Launch post-run tasks
            if self.monitor is not None:
                self.monitor.stop()
                self.monitor.close()
            self.post_run()

    def post_run(self):
        if not self.is_stdout_piped():
            wait("Forwarding {count} packets to wireshark".format(
                    count=str(self.monitor.packets_written)
                ),
                end=True
            )
            success("{count} packets have been forwarded to wireshark".format(
                    count = str(self.monitor.packets_written)
                )
            )
        super().post_run()


def wshark_main():
    app = WhadWiresharkApp()
    run_app(app)



"""
Exceptions related to RF4CE domain.
"""
class MissingRF4CEHeader(Exception):
    def __init__(self):
        super().__init__()

class MissingRF4CESecurityFlag(Exception):
    def __init__(self):
        super().__init__()

class MissingCryptographicMaterial(Exception):
    def __init__(self):
        super().__init__()



"""WHAD server tool

This utility implements a server module, allowing to create a TCP proxy
which can be used to access a device remotely.
"""
from prompt_toolkit import print_formatted_text, HTML

from whad.common.monitors.pcap import PcapWriterMonitor
from whad.cli.app import CommandLineApp, ApplicationError, run_app
from scapy.all import *
from whad.device.unix import  UnixSocketConnector, UnixConnector, UnixSocketServerDevice
from whad.device import Bridge, ProtocolHub
from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady
from whad.cli.ui import error, warning, success, info, display_event, display_packet, format_analyzer_output
from whad.tools.utils import get_analyzers

import logging
import time
import sys
import json

logger = logging.getLogger(__name__)
#logging.basicConfig(level=logging.DEBUG)

class WhadAnalyzeUnixSocketConnector(UnixSocketConnector):
    pass

class WhadAnalyzePipe(Bridge):

    def __init__(self, input_connector, output_connector, processing_function):
        super().__init__(input_connector, output_connector)
        self.processing_function = processing_function

    def on_outbound(self, message):
        """Process outbound messages.

        Outbound packets are packets coming from our input connector,that need to be
        forwarded as packets to the next tool.
        """
        if hasattr(message, "to_packet"):
            pkt = message.to_packet()
            pkts = self.processing_function(pkt, piped=True)
            if pkts is not None:
                for forwarded in pkts:
                    msg = message.from_packet(forwarded)
                    super().on_outbound(msg)
        else:
            logger.debug('[wfilter][input-pipe] forward default outbound message %s' % message)
            # Forward other messages
            super().on_outbound(message)


def display_analyzers(analyzers):
    for domain, analyzers_list in analyzers.items():
        print_formatted_text(HTML("<b><ansicyan>Available analyzers: </ansicyan> {domain}</b>".format(domain=domain)))
        for analyzer_name, analyzer_class in analyzers_list.items():
            print_formatted_text(HTML("  <b>- {analyzer_name}</b> : {output}".format(analyzer_name=analyzer_name, output=", ".join(analyzer_class().output.keys()))))

        print()

class WhadAnalyzeApp(CommandLineApp):

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD analyze tool',
            interface=False,
            commands=False,
            input=CommandLineApp.INPUT_WHAD,
            output=CommandLineApp.OUTPUT_STANDARD
        )

        self.add_argument(
            'analyzer',
            help='Analyzer to use',
            nargs="*"
        )


        self.add_argument(
            '--trigger',
            dest='trigger',
            default=False,
            action="store_true",
            help='show when an analyzer is triggered'
        )

        self.add_argument(
            '--json',
            dest='json',
            default=False,
            action="store_true",
            help='serialize output into json format'
        )

        self.add_argument(
            '-p',
            '--packets',
            dest='packets',
            default=False,
            action="store_true",
            help='display packets associated with the analyzer'
        )

        self.add_argument(
            '--label',
            dest='label',
            default=False,
            action="store_true",
            help='display labels before output value'
        )

        self.add_argument(
            '-d',
            '-D',
            '--delimiter',
            dest='delimiter',
            default="\n",
            help='delimiter between outputs'
        )

        self.add_argument(
            '-r',
            '--raw',
            dest='raw',
            action="store_true",
            default=False,
            help='dump output directly to stdout buffer'
        )

        self.add_argument(
            '-l',
            '--list',
            action="store_true",
            help='List of available analyzers'
        )

    def on_packet(self, pkt, piped=False):
        for analyzer_name, analyzer in self.selected_analyzers.items():
            analyzer.process_packet(pkt)

            if analyzer.triggered and not analyzer._displayed and self.args.trigger and not piped:
                info(analyzer_name + " → " + "triggered")
                analyzer._displayed = True

            if analyzer.completed:
                if self.args.packets and piped:
                    marked_packets = analyzer.marked_packets
                    analyzer.reset()
                    return marked_packets

                if analyzer_name in self.provided_parameters:
                    out = []
                    if not self.args.raw and not self.args.json:
                        for parameter in self.provided_parameters[analyzer_name]:
                            try:
                                out.append(
                                    ((parameter + ": ") if self.args.label else "") +
                                    format_analyzer_output(analyzer.output[parameter], mode="human_readable")
                                )
                            except KeyError:
                                error("Unknown output parameter {param}, ignoring.".format(param=parameter))
                        print(self.args.delimiter.join(out))
                    elif self.args.raw:
                        for parameter in self.provided_parameters[analyzer_name]:
                            try:
                                sys.stdout.buffer.write(format_analyzer_output(analyzer.output[parameter], mode="raw"))
                            except KeyError:
                                error("Unknown output parameter {param}, ignoring.".format(param=parameter))
                    elif self.args.json:
                        for parameter in self.provided_parameters[analyzer_name]:
                            try:
                                json_value = format_analyzer_output(analyzer.output[parameter], mode="json")
                                if json_value is not None:
                                    if self.args.label:
                                        print("{\"" + parameter + "\": " + json_value + "}")
                                    else:
                                        print(json_value)
                                else:
                                    error("Parameter {param} not serializable in JSON, ignoring.".format(param=parameter))
                            except KeyError:
                                error("Unknown output parameter {param}, ignoring.".format(param=parameter))

                    sys.stdout.flush()

                else:
                    if not self.args.json:
                        success("[✓] " + analyzer_name + " → " + "completed")
                        for output_name, output_value in analyzer.output.items():
                            print_formatted_text(
                                HTML("  - <b>{name}: </b> {value}".format(
                                        name=output_name,
                                        value=format_analyzer_output(output_value,mode="human_readable")
                                    )
                                )
                            )
                        if self.args.packets:
                            print()
                            print_formatted_text(HTML("  - <b>{count} packets analyzed: </b>".format(count=len(analyzer.marked_packets))))

                            for packet in analyzer.marked_packets:
                                display_packet(
                                    pkt,
                                    show_metadata = self.args.metadata,
                                    format = self.args.format
                                )
                        print()
                    else:
                            out = "{"
                            for output_name, output_value in analyzer.output.items():
                                json_value = format_analyzer_output(output_value,mode="json")
                                out += "\"" + output_name + "\": " + json_value + ", "
                            out = out[:-2] + "}"
                            print(out)
                    analyzer._displayed = False
                analyzer.reset()

    def get_provided_analyzers(self):
        available_analyzers = list(get_analyzers(self.args.domain).keys())
        analyzers = []
        parameters = {}
        for analyzer in self.args.analyzer:
            if "." in analyzer:
                name, param = analyzer.split(".")
                if name in available_analyzers:
                    analyzers.append(name)
                    if name in parameters:
                        parameters[name].append(param)
                    else:
                        parameters[name] = [param]
                else:
                    error("Unknown analyzer ({analyzer}).".format(analyzer=name))
                    exit(1)
            else:
                if analyzer in available_analyzers:
                    analyzers.append(analyzer)
                else:
                    error("Unknown analyzer ({analyzer}).".format(analyzer=analyzer))
                    exit(1)
        return list(set(analyzers)), parameters

    def run(self):
        # Launch pre-run tasks
        self.pre_run()

        if self.args.list:
            analyzers = get_analyzers()
            display_analyzers(analyzers)

        try:
            if self.is_piped_interface():
                interface = self.input_interface
            else:
                interface = self.interface

            if interface is not None:
                if not self.args.nocolor:
                    conf.color_theme = BrightTheme()

                parameters = self.args.__dict__

                connector = WhadAnalyzeUnixSocketConnector(interface)
                for parameter_name, parameter_value in parameters.items():
                    connector.add_parameter(parameter_name, parameter_value)

                self.provided_analyzers, self.provided_parameters = self.get_provided_analyzers()
                self.selected_analyzers = {}
                for analyzer_name, analyzer_class in get_analyzers(self.args.domain).items():
                    if analyzer_name in self.provided_analyzers or len(self.provided_analyzers) == 0:
                        self.selected_analyzers[analyzer_name] = analyzer_class()
                        self.selected_analyzers[analyzer_name]._displayed = False

                connector.domain = self.args.domain
                hub = ProtocolHub(2)
                connector.format = hub.get(self.args.domain).format


                if self.is_stdout_piped() and self.args.packets:
                    unix_server = UnixConnector(UnixSocketServerDevice(parameters={
                        'domain': self.args.domain,
                        'format': self.args.format,
                        'metadata' : self.args.metadata
                    }))


                    while not unix_server.device.opened:
                        if unix_server.device.timedout:
                            return
                        else:
                            sleep(0.1)
                    # Create our packet bridge
                    logger.info("[wanalyze] Starting our output pipe")
                    output_pipe = WhadAnalyzePipe(connector, unix_server, self.on_packet)
                else:
                    connector.on_packet = self.on_packet

                while interface.opened:
                    time.sleep(.1)

        except KeyboardInterrupt:
            # Launch post-run tasks
            self.post_run()


def wanalyze_main():
    app = WhadAnalyzeApp()
    run_app(app)



from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady, WhadDeviceAccessDenied
from whad.device.virtual import VirtualDevice
from whad.helpers import message_filter,is_message_type,bd_addr_to_bytes
from whad.hub.discovery import Domain, Capability

from whad.hub.generic.cmdresult import CommandResult
from whad.hub.dot15d4 import Commands

from whad.device.virtual.rzusbstick.constants import RZUSBStickInternalStates, \
    RZUSBStickId, RZUSBStickModes, RZUSBStickEndPoints, RZUSBStickCommands, \
    RZUSBStickResponses
from usb.core import find, USBError, USBTimeoutError
from usb.util import get_string
from struct import unpack, pack
from time import sleep

# Helpers functions
def get_rzusbstick(id=0,bus=None, address=None):
    '''
    Returns a RZUSBStick USB object based on index or bus and address.
    '''
    devices = list(find(idVendor=RZUSBStickId.RZUSBSTICK_ID_VENDOR, idProduct=RZUSBStickId.RZUSBSTICK_ID_PRODUCT,find_all=True))
    if bus is not None and address is not None:
        for device in devices:
            if device.bus == bus and device.address == address:
                return (devices.index(device), device)
        # No device found with the corresponding bus/address, return None
        return None
    else:
        try:
            return (id, devices[id])
        except IndexError:
            return None

class RZUSBStickDevice(VirtualDevice):

    INTERFACE_NAME = "rzusbstick"

    @classmethod
    def list(cls):
        '''
        Returns a list of available RZUSBStick devices.
        '''
        available_devices = []
        for rzusbstick in find(idVendor=RZUSBStickId.RZUSBSTICK_ID_VENDOR, idProduct=RZUSBStickId.RZUSBSTICK_ID_PRODUCT,find_all=True):
            available_devices.append(RZUSBStickDevice(bus=rzusbstick.bus, address=rzusbstick.address))
        return available_devices

    @property
    def identifier(self):
        '''
        Returns the identifier of the current device (e.g., bus + address in format "<bus>-<address>").
        '''
        return str(self.__rzusbstick.bus)+"-"+str(self.__rzusbstick.address)


    def __init__(self, index=0, bus=None, address=None):
        """
        Create device connection
        """
        device = get_rzusbstick(index,bus=bus,address=address)
        if device is None:
            raise WhadDeviceNotFound

        self.__opened = False
        self.__opened_stream = False
        self.__channel = 11
        self.__future_channel = 11
        self.__internal_state = RZUSBStickInternalStates.NONE
        self.__index, self.__rzusbstick = device
        super().__init__()

    def open(self):
        try:
            self.__rzusbstick.set_configuration()
        except USBError as err:
            if err.errno == 13:
                raise WhadDeviceAccessDenied("rzusbstick")
            else:
                raise WhadDeviceNotReady()
        self._dev_id = self._get_serial_number()
        self._fw_author = self._get_manufacturer()
        self._fw_url = self._get_url()
        self._fw_version = self._get_firmware_version()
        self._dev_capabilities = self._get_capabilities()

        self.__opened = True
        # Ask parent class to run a background I/O thread
        super().open()

    def write(self, data):
        if not self.__opened:
            raise WhadDeviceNotReady()

    def read(self):
        if not self.__opened:
            raise WhadDeviceNotReady()
        if self.__opened_stream:
            try:
                data = self.__rzusbstick_read_packet()
            except USBTimeoutError:
                data = b""
            if data is not None and len(data) >= 1:
                if data[0] == RZUSBStickResponses.RZ_AIRCAPTURE_DATA and len(data) >= 10:
                    self.__input_header = data[:9]
                    self.__input_buffer = data[9:]
                    self.__input_buffer_length = data[1] - 9
                elif len(self.__input_header) > 0:
                    self.__input_buffer += data
                else:
                    self.__input_buffer = b""
                    self.__input_buffer_length = 0
                    self.__input_header = b""

                if len(self.__input_header) == 9 and self.__input_buffer_length == len(self.__input_buffer):
                    rssi = 3 * self.__input_header[6] - 91
                    valid_fcs = (self.__input_header[7] == 0x01)
                    packet = self.__input_buffer[:-1]
                    link_quality_indicator = self.__input_buffer[-1]
                    self._send_whad_zigbee_raw_pdu(packet, rssi=rssi, is_fcs_valid=valid_fcs)

    def reset(self):
        self.__rzusbstick.reset()

    def close(self):
        super().close()

    # Virtual device whad message builder
    def _send_whad_zigbee_raw_pdu(self, packet, rssi=None, is_fcs_valid=None, timestamp=None):
        pdu = packet[:-2]
        fcs = unpack("H",packet[-2:])[0]

        # Create a RawPduReceived message
        msg = self.hub.dot15d4.create_raw_pdu_received(
            self.__channel,
            pdu,
            fcs,
            fcs_validity=is_fcs_valid
        )

        # Set timestamp and RSSI if provided
        if rssi is not None:
            msg.rssi = rssi
        if timestamp is not None:
            msg.timestamp = timestamp

        # Send message

        self._send_whad_message(msg)


    # Virtual device whad message callbacks
    def _on_whad_dot15d4_stop(self, message):
        if self._stop():
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_dot15d4_send_raw(self, message):
        channel = message.channel

        if self._set_channel(channel):
            packet = message.pdu + pack("H",message.fcs)
            success = self._send_packet(packet)
        else:
            success = False
        self._send_whad_command_result(CommandResult.SUCCESS if success else CommandResult.ERROR)

    def _on_whad_dot15d4_sniff(self, message):
        channel = message.channel
        self.__future_channel = channel
        self.__internal_state = RZUSBStickInternalStates.SNIFFING
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_dot15d4_start(self, message):
        self.__input_buffer = b""
        self.__input_buffer_length = 0
        self.__input_header = b""
        if self._start():
            if self.__future_channel != self.__channel:
                self._set_channel(self.__future_channel)
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.ERROR)

    # RZUSBStick low level communication primitives

    def __rzusbstick_read_packet(self, timeout=200):
        return bytes(self.__rzusbstick.read(RZUSBStickEndPoints.RZ_PACKET_ENDPOINT, self.__rzusbstick.bMaxPacketSize0, timeout=timeout))

    def _rzusbstick_read_response(self, timeout=200):
        return bytes(self.__rzusbstick.read(RZUSBStickEndPoints.RZ_RESPONSE_ENDPOINT, self.__rzusbstick.bMaxPacketSize0, timeout=timeout))

    def _rzusbstick_send_command(self, command, data=b"", timeout=200):
        data = [command] + list(data)
        self.__rzusbstick.write(RZUSBStickEndPoints.RZ_COMMAND_ENDPOINT, data, timeout=timeout)
        return self._rzusbstick_read_response()[0] == RZUSBStickResponses.RZ_RESP_SUCCESS

    # Discovery related functions
    def _get_capabilities(self):
        if "KILLERB" in self.__rzusbstick.product:
            capabilities = {
                Domain.Dot15d4 : (
                                    (Capability.Sniff | Capability.Inject),
                                    [Commands.Sniff, Commands.Send, Commands.Start, Commands.Stop]
                )
            }
        else:
            capabilities = {
                Domain.Dot15d4 : (
                                    (Capability.Sniff),
                                    [Commands.Sniff, Commands.Start, Commands.Stop]
                )
            }

        return capabilities

    def _get_manufacturer(self):
        if "KILLERB" in self.__rzusbstick.product:
            return "Joshua Wright (KillerBee version)".encode('utf-8')
        else:
            return (self.__rzusbstick.manufacturer + "(Factory version)").encode('utf-8')

    def _get_serial_number(self):
        return bytes.fromhex(
                                self.__rzusbstick.serial_number[:-1] +
                                "{:04x}".format(self.__rzusbstick.bus)  +
                                "{:04x}".format(self.__rzusbstick.address)
        )

    def _get_firmware_version(self):
        return (1, 0, 0)

    def _get_url(self):
        if "KILLERB" in self.__rzusbstick.product:
            return "https://github.com/riverloopsec/killerbee".encode('utf-8')
        else:
            return "https://www.microchip.com/en-us/development-tool/ATAVRRZUSBSTICK".encode('utf-8')

    # RZUSBStick commands

    def _stop(self):
        stream_closed = self._close_stream()
        stopped = self._set_mode(RZUSBStickModes.RZ_MODE_NONE)
        return stream_closed and stopped

    def _start(self):
        started = self._set_mode(RZUSBStickModes.RZ_MODE_AIRCAPTURE)
        stream_enabled = self._open_stream()
        return started and stream_enabled

    def _set_mode(self, mode=RZUSBStickModes.RZ_MODE_NONE):
        return self._rzusbstick_send_command(RZUSBStickCommands.RZ_SET_MODE, data=[mode])

    def _set_channel(self, channel=11):
        if channel < 11 or channel > 26:
            return False
        self._close_stream()
        success = self._rzusbstick_send_command(RZUSBStickCommands.RZ_SET_CHANNEL, data=[channel])
        if success:
            self.__channel = channel
        self._open_stream()
        return success

    def _get_channel(self):
        return self.__channel

    def _enable_jamming(self):
        return self._rzusbstick_send_command(RZUSBStickCommands.RZ_JAMMER_ON)

    def _disable_jamming(self):
        return self._rzusbstick_send_command(RZUSBStickCommands.RZ_JAMMER_OFF)

    def _open_stream(self):
        if not self.__opened_stream:
            success = self._rzusbstick_send_command(RZUSBStickCommands.RZ_OPEN_STREAM)
            self.__opened_stream = success
        return self.__opened_stream

    def _close_stream(self):
        if self.__opened_stream:
            success = self._rzusbstick_send_command(RZUSBStickCommands.RZ_CLOSE_STREAM)
            self.__opened_stream = not success
            return True
        return False

    def _send_packet(self, data):
        if len(data) >= 1 and len(data) <= 125:
            self._close_stream()
            success = self._rzusbstick_send_command(RZUSBStickCommands.RZ_INJECT_FRAME, bytes([len(data)])+data, timeout=700)
            self._open_stream()
            return success



"""
BLE GATT Characteristic Model
=============================
"""
from whad.ble.stack.att.constants import BleAttProperties, SecurityAccess
from whad.ble.profile.attribute import Attribute, UUID, get_uuid_alias
from whad.ble.exceptions import InvalidHandleValueException
from struct import pack, unpack

class CharacteristicProperties(object):
    BROADCAST = 0x01
    READ = 0x02
    WRITE_WITHOUT_RESPONSE = 0x04
    WRITE = 0x08
    NOTIFY = 0x10
    INDICATE = 0x20
    AUTH_SIGNED_WRITES = 0x40
    EXTENDED_PROPERTIES = 0x80

class CharacteristicDescriptor(Attribute):
    """BLE Characteristic descriptor
    """
    def __init__(self, characteristic, uuid, handle=0, value=b''):
        super().__init__(uuid=uuid,handle=handle,value=value)
        self.__characteristic = characteristic

    @property
    def characteristic(self):
        return self.__characteristic

    @property
    def uuid(self):
        return self.type_uuid
    
    @property
    def name(self):
        alias = get_uuid_alias(self.type_uuid)
        if alias is not None:
            return '%s (0x%s)' % (
                alias,
                str(self.type_uuid)
            )
        else:
            return str(self.type_uuid)

class ClientCharacteristicConfig(CharacteristicDescriptor):
    """Client Characteristic Configuration Descriptor
    """

    def __init__(self, characteristic, handle=0, notify=False, indicate=False):
        """Instanciate a Client Characteristic Configuration Descriptor

        :param bool notify: Set to True to get the corresponding characteristic notified on change
        :param bool indicate: Set to True to get the corresponding characteristic indicated on change
        """
        value = 0
        if notify:
            value |= 0x0001
        if indicate:
            value |= 0x0002
        super().__init__(characteristic, uuid=UUID(0x2902), handle=handle, value=pack('<H', value))

    @property
    def config(self):
        return unpack('<H', super().value)[0]

    @config.setter
    def config(self, val):
        super().value = pack('<H', val)


class ReportReferenceDescriptor(CharacteristicDescriptor):
    """Report Reference Descriptor, used in HID profile
    """

    def __init__(self, characteristic, handle=None, notify=False, indicate=False):
        """Instanciate a Report Reference Descriptor

        :param bool notify: Set to True to get the corresponding characteristic notified on change
        :param bool indicate: Set to True to get the corresponding characteristic indicated on change
        """
        super().__init__(
            characteristic,
            uuid=UUID(0x2908),
            handle=handle,
            value=b'\x01\x01'
        )

class CharacteristicUserDescriptionDescriptor(CharacteristicDescriptor):
    """Characteristic description defined by user, contains
    a textual description of the related characteristic.
    """
    def __init__(self, characteristic, handle=None, description=''):
        """Instanciate a Characteristic User Description descriptor

        :param description: Set characteristic text description
        """
        super().__init__(
            characteristic,
            uuid=UUID(0x2901),
            handle=handle,
            value=description.encode('utf-8')
        )


class CharacteristicValue(Attribute):
    def __init__(self, uuid, handle=None, value=b'', characteristic=None):
        super().__init__(uuid=uuid, handle=handle, value=value)
        self.__characteristic = characteristic

    @property
    def characteristic(self):
        return self.__characteristic

    @property
    def uuid(self):
        return self.type_uuid


class Characteristic(Attribute):
    """BLE Characteristic
    """

    def __init__(self, uuid, handle=0, end_handle=0, value=b'', properties=BleAttProperties.DEFAULT, description=None, security=[]):
        """Instanciate a BLE characteristic object

        :param uuid: 16-bit or 128-bit UUID
        :param int handle: Handle value
        :param bytes value: Characteristic value
        :param int perms: Permissions
        """
        super().__init__(uuid=UUID(0x2803), handle=handle,
                         value=pack('<BH', properties & 0xff, handle+1)+uuid.to_bytes())
        self.__handle = handle
        if end_handle == 0:
            self.__end_handle = handle
        else:
            self.__end_handle = end_handle
        self.__charac_uuid = uuid

        # notification and indication callbacks
        self.__notification_callback = None
        self.__indication_callback = None

        if isinstance(handle, int):
            self.__value_handle = handle + 1
            self.__value = CharacteristicValue(uuid, self.__value_handle, value, self)
            self.__end_handle = self.__value_handle
        else:
            self.__value_handle = None
            self.__value = CharacteristicValue(uuid, self.__value_handle, value, self)

        # Permissions
        self.__properties = properties

        # Security properties
        self.__security = security

        # Descriptors
        self.__descriptors = []

    def payload(self):
        """Return characteristic payload
        """
        #return bytes([self.__properties, self.__value_handle]) + self.__value.uuid.to_bytes()
        return pack('<BH', self.__properties, self.__value_handle) + self.__value.uuid.packed

    def set_notification_callback(self, callback):
        self.__notification_callback = callback

    def set_indication_callback(self, callback):
        self.__indication_callback = callback

    @Attribute.handle.setter
    def handle(self, new_handle):
        if isinstance(new_handle, int):

            # Set attribute handle
            Attribute.handle.fset(self, new_handle)

            self.__value.handle = self.handle + 1
            self.__value_handle = self.handle + 1

            handle = self.__value_handle

            # Update descriptors handle
            for descriptor in self.__descriptors:
                handle += 1
                descriptor.handle = handle

            # Update end handle
            self.__end_handle = handle
        else:
            raise InvalidHandleValueException

    @property
    def value(self):
        return self.__value.value

    @property
    def value_attr(self):
        return self.__value

    @property
    def value_handle(self):
        return self.__value_handle

    @value_handle.setter
    def value_handle(self, value):
        self.__value_handle = value
        self.__value.handle = value

    @value.setter
    def value(self, new_value):
        """Set characteristic value

        :param bytes new_value: Value
        """
        self.__value.value = new_value

        # Notify or indicate if required
        if self.must_notify() and self.__notification_callback is not None:
            self.__notification_callback(self)
        elif self.must_indicate() and self.__indication_callback is not None:
            self.__indication_callback(self)

    @property
    def properties(self):
        return self.__properties

    @properties.setter
    def properties(self, new_properties):
        self.__properties = new_properties

    @property
    def uuid(self):
        return self.__charac_uuid

    @uuid.setter
    def uuid(self, value):
        self.__charac_uuid = value

    @property
    def end_handle(self):
        return self.__end_handle

    @property
    def name(self):
        alias = get_uuid_alias(self.__charac_uuid)
        if alias is not None:
            return '%s (0x%s)' % (
                alias,
                str(self.__charac_uuid)
            )
        else:
            return str(self.__charac_uuid)

    ##########################
    # Methods
    ##########################

    def readable(self):
        return (self.properties & CharacteristicProperties.READ) != 0

    def writeable(self):
        return (
            ((self.properties & CharacteristicProperties.WRITE) != 0) or
            ((self.properties & CharacteristicProperties.WRITE_WITHOUT_RESPONSE) != 0)
        )

    def can_notify(self) -> bool:
        """Determine if characteristic sends notifications.

        :return: ``True`` if characteristic sends notification, ``False`` otherwise.
        :rtype: bool
        """
        return ((self.properties & CharacteristicProperties.NOTIFY) != 0)

    def must_notify(self):
        """Determine if a notification must be sent for this characteristic.

        Notification must be sent when a characteristic has the notification property and
        its ClientCharacteristicConfiguration descriptor has notifications enabled.
        """
        if (self.properties & CharacteristicProperties.NOTIFY) != 0:
            cccd = self.get_client_config()
            if cccd is not None:
                return (cccd.config == 0x0001)
        return False

    def can_indicate(self) -> bool:
        """Determine if characteristic sends indications.

        :return: ``True`` if characteristic sends indication, ``False`` otherwise.
        :rtype: bool
        """
        return ((self.properties & CharacteristicProperties.INDICATE) != 0)

    def must_indicate(self):
        """Determine if an indication must be sent for this characteristic.

        Indication must be sent when a characteristic has the indication property and
        its ClientCharacteristicConfiguration descriptor has indications enabled.
        """
        if (self.properties & CharacteristicProperties.INDICATE) != 0:
            cccd = self.get_client_config()
            if cccd is not None:
                return (cccd.config == 0x0002)
        return False



    def add_descriptor(self, descriptor):
        """Add a descriptor

        :param CharacteristicDescriptor descriptor: Descriptor instance to add to this characteristic.
        """
        if isinstance(descriptor, CharacteristicDescriptor):
            self.__descriptors.append(descriptor)
            if descriptor.handle > self.__end_handle:
                self.__end_handle = descriptor.handle

    def descriptors(self):
        """Iterate over the registered descriptors (generator)
        """
        for desc in self.__descriptors:
            yield desc

    def get_client_config(self):
        """Find characteristic client configuration descriptor
        """
        for desc in self.__descriptors:
            if isinstance(desc, ClientCharacteristicConfig):
                return desc
        return None

    @property
    def security(self) -> SecurityAccess:
        """Returns security access property
        """
        return self.__security

    def get_security_access(self, access_type):
        """Returns the security access properties linked to an access type.

        :param access_type: access type to check
        :type access_type: SecurityAccess
        """
        for access in self.__security:
            if isinstance(access, access_type):
                return access
        return None

    def check_security_property(self, access_type, property):
        """Returns a boolean indicating if a property is required for a given access type.

        :param access_type: access type to check
        :type access_type: SecurityAccess
        :param property: security property to check
        :type property: SecurityProperty
        """
        access = self.get_security_access(access_type)
        if access is not None:
            return property in access.access
        return False



"""This module provides different classes that represent a BLE device and
allows to interact with it:

* :class:`whad.ble.profile.GenericProfile` is a base class used to register all the ATT attributes, including
  services, characteristics, characteristic values and descriptors. It is able to
  inspect any derived class and build the corresponding profile based on properties
  declared with :class:`whad.ble.profile.service.PrimaryService` and
  :class:`whad.ble.profile.characteristic.Characteristic`.

"""
import json
from typing import List, Iterator
from whad.ble.stack.smp import Pairing
from whad.ble.profile.attribute import Attribute, UUID
from whad.ble.profile.characteristic import Characteristic as BleCharacteristic,\
    CharacteristicProperties, ClientCharacteristicConfig, \
    CharacteristicValue as BleCharacteristicValue, \
    CharacteristicDescriptor as BleCharacteristicDescriptor, \
    ReportReferenceDescriptor as BleReportReferenceDescriptor, \
    CharacteristicUserDescriptionDescriptor as BleCharacteristicUserDescriptionDescriptor

from whad.ble.profile.service import PrimaryService as BlePrimaryService, \
    SecondaryService as BleSecondaryService, IncludeService as BleIncludeService, \
    Service
from whad.ble.exceptions import InvalidHandleValueException
from whad.ble.stack.att.constants import BleAttProperties, SecurityProperty, \
    SecurityAccess, ReadAccess, WriteAccess, Authentication, Authorization, Encryption

import logging
logger = logging.getLogger(__name__)

###################################
# Decorators for GenericProfile
###################################

class CharacteristicHook(object):

    def __init__(self, *args):
        if len(args) == 1:
            characteristic = args[0]
            if isinstance(characteristic, Characteristic):
                self.__characteristic = str(characteristic.service.uuid)+':'+str(characteristic.uuid)
            else:
                raise TypeError
        elif len(args) == 2:
            service = args[0]
            charac = args[1]
            if isinstance(service, str) and isinstance(charac, str):
                self.__characteristic = str(UUID(service)) + ':' + str(UUID(charac))
            elif isinstance(service, UUID) and isinstance(charac, UUID):
                self.__characteristic = str(service) + ':' + str(charac)

    def __call__(self, method):
        if not hasattr(method, 'hooks'):
            method.hooks = []
        if not hasattr(method, 'characteristic'):
            method.characteristic = self.__characteristic
        return method

class read(CharacteristicHook):
    """Read hook decorator

    This decorator is used to declare a callback method for read operations
    on a specific characteristic.
    """

    def __call__(self, method):
        """Add a specific hook to the method
        """
        super().__call__(method)
        if 'read' not in method.hooks:
            method.hooks.append('read')
        return method

class write(CharacteristicHook):

    def __call__(self, method):
        """Add a specific hook to the method
        """
        super().__call__(method)
        if 'write' not in method.hooks:
            method.hooks.append('write')
        return method

class written(CharacteristicHook):

    def __call__(self, method):
        """Add a specific hook to the method
        """
        super().__call__(method)
        if 'written' not in method.hooks:
            method.hooks.append('written')
        return method

class subscribed(CharacteristicHook):

    def __call__(self, method):
        """Add a specific hook to the method
        """
        super().__call__(method)
        if 'sub' not in method.hooks:
            method.hooks.append('sub')
        return method

class unsubscribed(CharacteristicHook):

    def __call__(self, method):
        """Add a specific hook to the method
        """
        super().__call__(method)
        if 'unsub' not in method.hooks:
            method.hooks.append('unsub')
        return method

def is_method_hook(method):
    """Determine if a method is a characteristic operation hook
    """
    if hasattr(method, 'hooks') and hasattr(method, 'characteristic'):
        return (len(method.hooks) > 0)
    return False

################################
# Descriptors model
#
# This section contains all the descriptor models to use while creating
# a profile from Python code. It contains a set of alternative classes
# used by GenericProfile to build the attribute database and populate an
# instance with the corresponding properties and objects.
################################

class CharacteristicDescriptor(object):
    """Generic CharacteristicDescriptor model
    """
    def __init__(self, bleclass=None):
        """Instanciate a characteristic descriptor model

        :param str name: attribute name to access this descriptor
        :param class bleclass: BLE descriptor class to use when instanciating the model
        :param list permissions: descriptor permissions (read/write/notify/indicate)
        """
        self.__handle = 0
        self.__class = bleclass

    @property
    def handle(self):
        return self.__handle

    @handle.setter
    def handle(self, value):
        self.__handle = value

    @property
    def bleclass(self):
        return self.__class


class ReportReferenceDescriptor(CharacteristicDescriptor):
    """Report Reference Descriptor model
    """
    def __init__(self, permissions=None):
        super().__init__(BleReportReferenceDescriptor)

class UserDescriptionDescriptor(CharacteristicDescriptor):
    """User description model
    """
    def __init__(self, description=''):
        super().__init__(BleCharacteristicUserDescriptionDescriptor)


class Characteristic(object):
    """GATT characteristic.
    """
    def __init__(self, name=None, uuid=None, value=b'', permissions=None, notify=False, indicate=False, description=None, security = [], **kwargs):
        """Declares a GATT characteristic.

        Other named arguments are used to declare characteristic's descriptors.

        :param  name:           Characteristic name used in GATT model
        :type   name:           str
        :param  uuid:           Characteristic UUID
        :type   uuid:           :class:`whad.ble.profile.attribute.UUID`
        :param  permissions:    List of permissions for this characteristic (*read*, *write*, *notify*, *indicate*)
        :type   permissions:    list
        :param  notify:         Enable notifications
        :type   notify:         bool
        :param  indicate:       Enable indications
        :type   indicate:       bool
        :param  description:    Textual description for this characteristic
        :type   description:    str
        :param security:        Indicate the security property associated to this characteristic
        :type security:         SecurityAccess
        """
        self.__handle = 0
        self.__name = name
        self.__uuid = uuid
        self.__value = value
        self.__perms = permissions
        self.__notify = notify
        self.__indicate = indicate
        self.__security = SecurityAccess.generate(security)
        self.__service = None
        self.__description = description
        self.__descriptors = []

        # Loop on kwargs to find descriptors
        for arg in kwargs:
            if isinstance(kwargs[arg], CharacteristicDescriptor):
                descriptor = kwargs[arg]
                descriptor.handle = 0
                descriptor.name = arg
                self.add_descriptor(descriptor)

                # Add descriptor to a property to this ServiceModel instance
                if not hasattr(self, arg):
                    setattr(self, arg, descriptor)

    def add_descriptor(self, descriptor):
        """Add descriptor to our descriptor list

        :param  descriptor: Descriptor to add to the characteristic's descriptor list
        :type   descriptor: :class:`whad.ble.profile.characteristic.CharacteristicDescriptor`
        """
        self.__descriptors.append(descriptor)

    def descriptors(self) -> Iterator[CharacteristicDescriptor]:
        """Enumerate descriptors attached to this characteristic

        This method will yield every descriptor attached to the characteristic.
        """
        for descriptor in self.__descriptors:
            yield descriptor

    def get_required_handles(self) -> int:
        """Compute the number of handles this characteristic will consume

        :return: Number of handles
        :rtype: int
        """
        handles = 2
        # A more handle as we may need a ClientCharacteristicConfiguration descriptor
        if self.__notify or self.__indicate:
            handles += 1
        return handles

    def attach(self, service):
        """Attach this characteristic to the corresponding service.

        :param  service:    Service
        :type   service:    :class:̀ whad.ble.profile.service.Service`
        """
        self.__service = service

    @property
    def handle(self) -> int:
        """Characteristic handle
        """
        return self.__handle

    @handle.setter
    def handle(self, value):
        """Set characteristic handle.

        :param  value:  New handle value
        :type   value:  int
        """
        self.__handle = value

    @property
    def end_handle(self) -> int:
        """Characteristic end handle (including characteristic value and descriptors).
        """
        return self.handle + self.get_required_handles()

    @property
    def name(self) -> str:
        """Name
        """
        return self.__name

    @name.setter
    def name(self, value):
        """Set characteristic name.

        :param  value:  New name
        :type   value:  str
        """
        self.__name = value

    @property
    def uuid(self):
        return self.__uuid

    @property
    def value(self) -> UUID:
        """Characteristic UUID
        """
        return self.__value

    @property
    def permissions(self) -> List[str]:
        """Characteristics permissions
        """
        return self.__perms

    @property
    def must_notify(self) -> bool:
        """Check if notification has to be sent on value change.
        """
        return self.__notify

    @property
    def must_indicate(self) -> bool:
        """Check if indication has to be sent on value change.
        """
        return self.__indicate

    @property
    def description(self) -> str:
        """Return characteristic textual description, if any
        """
        return self.__description

    @property
    def service(self) -> Service:
        """Related service.
        """
        return self.__service

    @property
    def security(self) -> SecurityAccess:
        """Returns security access property
        """
        return self.__security

class ServiceModel(object):

    PRIMARY = 1
    SECONDARY = 2

    def __init__(self, uuid=None, start_handle=None, end_handle=None, name=None, service_type=PRIMARY, **kwargs):
        self.__handle = 0
        self.__end_handle = 0
        self.__uuid = uuid
        self.__name = name
        self.__characteristics = []
        self.__included_services = []

        if start_handle is None:
            self.__handle = 0
        else:
            self.__handle = start_handle

        if end_handle is None:
            self.__end_handle = 0
        else:
            self.__end_handle = end_handle

        # Loop on kwargs to find characteristics and included services
        for arg in kwargs:
            if isinstance(kwargs[arg], Characteristic):
                charac = kwargs[arg]
                charac.handle = 0
                charac.name = arg
                self.add_characteristic(charac)
                charac.attach(self)

                # Add characteristic to a property to this ServiceModel instance
                if not hasattr(self, arg):
                    setattr(self, arg, charac)
            elif isinstance(kwargs[arg], SecondaryService):
                # We must include this secondary service in this service
                service = kwargs[arg]
                self.add_included_service(service)

                # Add included service to a property to this ServiceModel instance
                if not hasattr(self, arg):
                    setattr(self, arg, service)


    def add_characteristic(self, characteristic_model):
        """Add a characteristic to the model
        """
        # Add characteristic to the list of our characteristics
        self.__characteristics.append(characteristic_model)

        # Update end handle value (include definition is a single attribute)
        if characteristic_model.handle >= self.__end_handle:
            self.__end_handle = characteristic_model.handle

    def add_included_service(self, service_model):
        """Add an included service to the model
        """
        self.__included_services.append(service_model)

        # Update end handle value
        if service_model.end >= self.__end_handle:
            self.__end_handle = service_model.end

    @property
    def uuid(self):
        return self.__uuid

    @property
    def handle(self):
        return self.__handle

    @property
    def end(self):
        return self.__end_handle

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, value):
        self.__name = value

    @handle.setter
    def handle(self, value):
        self.__handle = value

    def characteristics(self):
        for charac in self.__characteristics:
            yield charac

    def included_services(self):
        for inc_service in self.__included_services:
            yield inc_service

class PrimaryService(ServiceModel):
    def __init__(self, uuid=None, start_handle=0, end_handle=0, name=None, **kwargs):
        """Declares a GATT primary service.

        Other named arguments are used to add service's characteristics.

        :param  uuid:           Primary service UUID
        :type   uuid:           :class:`whad.ble.profile.attribute.UUID`
        :param  start_handle:   Service start handle
        :type   start_handle:   int, optional
        :param  end_handle:     Service end handle
        :type   end_handle:     int, optional
        :param  name:           Service name
        :type   name:           str
        """
        super().__init__(uuid, start_handle, end_handle, service_type=ServiceModel.PRIMARY, name=name, **kwargs)


class SecondaryService(ServiceModel):
    def __init__(self, uuid=None, start_handle=0, end_handle=0, name=None, **kwargs):
        """Declares a GATT secondary service.

        Other named arguments are used to add service's characteristics.

        :param  uuid:           Primary service UUID
        :type   uuid:           :class:`whad.ble.profile.attribute.UUID`
        :param  start_handle:   Service start handle
        :type   start_handle:   int, optional
        :param  end_handle:     Service end handle
        :type   end_handle:     int, optional
        :param  name:           Service name
        :type   name:           str
        """
        super().__init__(uuid, start_handle, end_handle, service_type=ServiceModel.SECONDARY, name=name, **kwargs)


class GenericProfile(object):
    """Generic Profile
    """

    def __init__(self, start_handle=0, from_json=None):
        """Parse the device model, instanciate all the services, characteristics
        and descriptors, compute all handle values and registers everything
        inside this instance for further use.

        :param  start_handle:   Start handle value to use (default: 0)
        :type   start_handle:   int
        :param  from_json:      JSON data describing a GATT profile
        :type   from_json:      str
        """
        self.__attr_db = {}
        self.__services = []
        self.__service_by_characteristic_handle = {}

        self.__start_handle = start_handle
        self.__handle = self.__start_handle

        self.__hooks = {}

        # Populate attribute database and model from JSON export if provided
        if from_json is not None:
            from_json = json.loads(from_json)
            # Parse JSON services, characteristics and descriptors to create
            # the corresponding model and attribute database
            if 'services' in from_json:
                # Loop on services
                for service in from_json['services']:
                    # Collect characteristics
                    service_characs = []
                    service_last_handle = service['start_handle']
                    if UUID(service['type_uuid']) == UUID(0x2800):
                        service_obj = BlePrimaryService(
                            uuid=UUID(service['uuid']),
                            handle=service['start_handle']
                        )
                    elif UUID(service['type_uuid']) == UUID(0x2801):
                        service_obj = BleSecondaryService(
                            uuid=UUID(service['uuid']),
                            handle=service['start_handle']
                        )
                    else:
                        # This is not a known service type UUID, continue with
                        # next service
                        continue

                    if 'characteristics' in service:
                        for charac in service['characteristics']:
                            charac_obj = BleCharacteristic(
                                uuid=UUID(charac['value']['uuid']),
                                handle=charac['handle'],
                                value=b'',
                                properties=charac['properties'],
                                security=SecurityAccess.int_to_accesses(charac['security'])
                            )

                            # Loop on descriptors, only support CCC at the moment
                            for desc in charac['descriptors']:
                                if UUID(desc['uuid']) == UUID(0x2902):
                                    desc_obj = ClientCharacteristicConfig(
                                        charac_obj,
                                        handle=desc['handle'],
                                        notify=charac_obj.must_notify(),
                                        indicate=charac_obj.must_indicate()
                                    )
                                    charac_obj.add_descriptor(desc_obj)
                                    self.register_attribute(desc_obj)

                            # Register characteristic and its value
                            self.register_attribute(charac_obj)
                            self.register_attribute(charac_obj.value_attr)

                            # Add characteristic to its related service
                            service_obj.add_characteristic(charac_obj)

                    self.register_attribute(service_obj)
                    self.add_service(service_obj)
        else:
            # Introspect this class definition and build model
            services = []
            props = dir(self)
            for prop in props:
                if not prop.startswith('_'):
                    if isinstance(getattr(self, prop), ServiceModel):
                        service = getattr(self, prop)
                        service.name = prop
                        services.append(service)


            # Instanciate each service, and for each of them the corresponding
            # characteristics
            for service in services:
                if isinstance(service, PrimaryService):
                    logger.info('creating primary service %s' % service.uuid)
                    # Create service
                    service_obj = BlePrimaryService(
                        uuid=service.uuid,
                        handle=self.__alloc_handle()
                    )

                elif isinstance(service, SecondaryService):
                    logger.info('creating secondary service %s' % service.uuid)
                    # Create service
                    service_obj = BleSecondaryService(
                        uuid=service.uuid,
                        handle=self.__alloc_handle()
                    )
                    self.__attr_db[service_obj.handle] = service_obj
                else:
                    continue

                # Create the corresponding instance property
                setattr(self, service.name, service_obj)

                # Loop on included services and create them if required
                for inc_service in service.included_services():
                    inc_service_obj = BleIncludeService(
                        uuid=inc_service.uuid,
                        handle=self.__alloc_handle(1),
                        start_handle=inc_service.handle,
                        end_handle=inc_service.end
                    )
                    self.__handle = inc_service_obj.end_handle

                    # Register this service include definition
                    self.register_attribute(inc_service_obj)
                    service_obj.add_include_service(inc_service_obj)

                # Loop on underlying characteristics, and create them too.
                for charac in service.characteristics():
                    charac_props = 0
                    if 'read' in charac.permissions:
                        charac_props |= CharacteristicProperties.READ
                    if 'write' in charac.permissions:
                        charac_props |= CharacteristicProperties.WRITE
                    if 'write_without_response' in charac.permissions:
                        charac_props |= CharacteristicProperties.WRITE_WITHOUT_RESPONSE
                    if charac.must_notify:
                        charac_props |= CharacteristicProperties.NOTIFY
                    if charac.must_indicate:
                        charac_props |= CharacteristicProperties.INDICATE
                    charac_obj = BleCharacteristic(
                        uuid=charac.uuid,
                        handle=self.__alloc_handle(1),
                        value=charac.value,
                        properties=charac_props,
                        security=charac.security
                    )
                    logger.info(' creating characteristic %s (handle:%d)' % (
                        charac_obj.uuid, charac_obj.handle
                    ))
                    self.__handle = charac_obj.end_handle

                    # Register this characteristic
                    self.register_attribute(charac_obj)
                    self.register_attribute(charac_obj.value_attr)

                    # If notify or indicate is set to true, we must add a new CCC descriptor
                    if charac.must_notify or charac.must_indicate:
                        ccc_desc = ClientCharacteristicConfig(
                            charac_obj,
                            handle=self.__alloc_handle(),
                            notify=charac.must_notify,
                            indicate=charac.must_indicate
                        )
                        logger.info('  creating cccd (handle:%d)' % (
                            ccc_desc.handle
                        ))
                        charac_obj.add_descriptor(ccc_desc)
                        self.register_attribute(ccc_desc)

                    # If characteristic description has been set, add a descriptor
                    if charac.description is not None:
                        cudd_desc = BleCharacteristicUserDescriptionDescriptor(
                            charac_obj,
                            handle=self.__alloc_handle(),
                            description=charac.description
                        )
                        logger.info('  creating cudd (handle:%d) with text "%s"' % (
                            cudd_desc.handle,
                            charac.description
                        ))
                        charac_obj.add_descriptor(cudd_desc)
                        self.register_attribute(cudd_desc)


                    # Loop on other characteristic descriptors and add them
                    for descriptor in charac.descriptors():
                        desc = descriptor.bleclass(
                            charac_obj,
                            handle=self.__alloc_handle()
                        )
                        logger.info('  creating %s descriptor (handle:%d)' % (
                            type(desc),
                            desc.handle
                        ))
                        charac_obj.add_descriptor(desc)
                        self.register_attribute(desc)

                    # Add our characteristic object to the corresponding service
                    setattr(service_obj, charac.name, charac_obj)
                    service_obj.add_characteristic(charac_obj)

                self.add_service(service_obj)

            # We then need to update included service start and end handles
            for inc_service in self.included_services():
                # Retrieve the included service UUID
                service_uuid = inc_service.service_uuid
                
                # Retrieve the corresponding object by UUID
                service_obj = self.get_service_by_UUID(service_uuid)

                # If found, update start and end handles
                if service_obj is not None:
                    inc_service.service_start_handle = service_obj.handle
                    inc_service.service_end_handle = service_obj.end_handle

        # Register any hook function
        props = dir(self)
        for prop in props:
            prop_obj = getattr(self, prop)
            # Is this property a callable hook ?
            if callable(prop_obj) and is_method_hook(prop_obj):
                # Associate hook method with each operation
                if prop_obj.characteristic not in self.__hooks:
                    self.__hooks[prop_obj.characteristic] = {}
                for operation in prop_obj.hooks:
                    self.__hooks[prop_obj.characteristic][operation] = prop_obj


    def __alloc_handle(self, number=1):
        """Allocate one or more handle values.

        :param  number: Number of handle values to allocate
        :type   number: int

        :return: Current handle value
        :rtype: int
        """
        self.__handle += number
        return self.__handle

    def __repr__(self):
        output = ''
        for service in self.services():
            output += 'Service %s (handles from %d to %d):\n' % (
                service.uuid,
                service.handle,
                service.end_handle
            )

            for inc_service in service.included_services():
                output += '  Included service %s (handle:%d, start_handle:%d, end_handle:%d)\n' % (inc_service.service_uuid, inc_service.handle, inc_service.service_start_handle, inc_service.service_end_handle)
            for charac in service.characteristics():
                properties = charac.properties
                charac_rights = ''
                if properties & CharacteristicProperties.READ != 0:
                    charac_rights += 'R'
                if properties & CharacteristicProperties.WRITE != 0:
                    charac_rights += 'W'
                if properties & CharacteristicProperties.INDICATE != 0:
                    charac_rights += 'I'
                if properties & CharacteristicProperties.NOTIFY != 0:
                    charac_rights += 'N'

                output += '  Characteristic %s (handle:%d, value handle: %d, props: %s)\n' % (
                    charac.uuid,
                    charac.handle,
                    charac.value_handle,
                    charac_rights
                )
                for desc in charac.descriptors():
                    output += '    Descriptor %s (handle: %d)\n' % (
                        desc.type_uuid,
                        desc.handle
                    )
        return output

    def register_attribute(self, attribute):
        """Register a GATT attribute

        :param  attribute:  Attribute to register
        :type   attribute:  :class:`whad.ble.profile.attribute.Attribute`
        """
        if isinstance(attribute, Attribute):
            self.__attr_db[attribute.handle] = attribute


    def add_service(self, service, handles_only=False):
        """Add a service to the current device

        :param  service:        Service to add to the device
        :type   service:        :class:`whad.ble.profile.service.Service`
        :param  handles_only:   Add only service handles if set to ``True``
        :type   handles_only:   bool
        """
        logger.debug('add service %s' % service.uuid)
        if service.handle == 0:
            # Service has not been fully configured, update its handle
            # and the handles of its characteristics and descriptors.
            service.handle = self.__alloc_handle()

        # Append service to the list of our services
        if not handles_only:
            self.__services.append(service)

        # Register service as an attribute
        self.register_attribute(service)

        # Register all its characteristics
        for charac in service.characteristics():
            # Register Characteristic and its CharacteristicValue
            self.register_attribute(charac)
            self.register_attribute(charac.value_attr)

            # Register characteristic's descriptors
            for desc in charac.descriptors():
                self.register_attribute(desc)

            # Add characteristic in our lookup table
            self.__service_by_characteristic_handle[charac.handle] = service

        # Update our last handle based on service's end handle
        self.__handle = service.end_handle


    def remove_service(self, service, handles_only=False):
        """Remove service

        :param  service:        Service object or UUID
        :type   service:        :class:`whad.ble.profile.service.Service`
        :param  handles_only:   Remove only handles if set to ``True``
        :type   handles_only:   bool
        """
        if isinstance(service, BlePrimaryService) or isinstance(service, BleSecondaryService):
            service_obj = self.get_service_by_UUID(service.uuid)
        elif isinstance(service, UUID):
            service_obj = self.get_service_by_UUID(service)
        else:
            service_obj = None

        # Process service object
        if service_obj is not None:
            # Remove service and all its characteristics from the attribute DB
            for charac in service_obj.characteristics():
                # Remove characteristic handle
                if charac.handle in self.__attr_db:
                    del self.__attr_db[charac.handle]

                # Remove characteristic value handle
                if charac.value_handle in self.__attr_db:
                    del self.__attr_db[charac.value_handle]

                # Remove all the attached descriptors
                for desc in charac.descriptors():
                    if desc.handle in self.__attr_db:
                        del self.__attr_db[desc.handle]

            # Remove service object from attribute db
            del self.__attr_db[service_obj.handle]

            # Remove service from our list of services (if required)
            if not handles_only:
                self.__services.remove(service)
        else:
            # Not found, raise IndexError
            raise IndexError()


    def update_service(self, service) -> bool:
        """Update service in profile.

        Keep service in place in the service list,
        but update all the services declared after this one.

        :param  service:    Service object to update.
        :type   service:    :class:`whad.ble.profile.service.Service`
        :return: ``True`` if service has been updated, ``False`` otherwise.
        :rtype: bool
        """
        try:
            service_index = self.__services.index(service)

            # Remove all handles used by this service
            self.remove_service(service, handles_only=True)

            # Register all the handles back into our attribute DB
            self.add_service(service, handles_only=True)

            # Update all other services
            handle = service.end_handle
            for remaining_service in self.__services[service_index+1:]:
                remaining_service.handle = handle + 1
                self.update_service(remaining_service)
                handle = remaining_service.end_handle
            self.__handle = handle
            return True
        except IndexError as notfound:
            return False

    def find_object_by_handle(self, handle) -> Attribute:
        """Find an object by its handle value

        :param  handle: Object handle
        :type   handle: int
        :return: Object if handle is valid, or raise an IndexError exception otherwise
        :rtype: :class:`whad.ble.profile.attribute.Attribute`
        :raises: IndexError
        """
        if handle in self.__attr_db:
            return self.__attr_db[handle]
        else:
            raise IndexError

    def find_objects_by_range(self, start, end) -> List[Attribute]:
        """Find attributes with handles belonging in the [start, end+1] interval.

        :param  start:  Start handle value
        :type   start:  int
        :param  end:    End handle value
        :type   end:    int
        :return:        List of objects with handles between start and end values
        :rtype: list
        """
        handles = []
        for handle in self.__attr_db:
            if handle>=start and handle<=end:
                handles.append(handle)
        handles.sort()
        return [self.find_object_by_handle(handle) for handle in handles]


    def find_characteristic_by_value_handle(self, value_handle) -> BleCharacteristic:
        """Find characteristic object by its value handle.

        :param  value_handle:   Characteristic value handle
        :type   value_handle:   int
        :return: Corresponding characteristic object or ``None`` if not found.
        :rtype: :class:`whad.ble.profile.characteristic.Characteristic`
        """
        try:
            char_value = self.find_object_by_handle(value_handle)
            if char_value is not None and hasattr(char_value, 'characteristic'):
                return char_value.characteristic
            else:
                return None
        except InvalidHandleValueException as bad_handle:
            return None


    def find_characteristic_end_handle(self, handle) -> int:
        """Find characteristic end handle based on its handle.

        :param  handle: Characteristic handle
        :type   handle: int
        :rtype: int
        :return: Characteristic value handle
        :raises: :class:`whad.ble.exceptions.InvalidHandleValueException`
        """
        try:
            # Find service owning the characteristic
            service = self.find_service_by_characteristic_handle(handle)

            # Build a list of characteristic handles
            service_char_handles=[]
            for characteristic in service.characteristics():
                service_char_handles.append(characteristic.handle)

            # Sort handles
            service_char_handles.sort()
            idx = service_char_handles.index(handle)
            if idx == len(service_char_handles) - 1:
                return service.end_handle
            else:
                return (service_char_handles[idx+1] - 1)

        except InvalidHandleValueException:
            return None


    def find_service_by_characteristic_handle(self, handle) -> Service:
        """Find a service object given a characteristic handle that belongs
        to this service.

        :param  handle: Characteristic handle belonging to the searched service
        :type   handle: int
        :rtype: :class:`whad.ble.profile.service.Service`
        :return: Service object containing the specified characteristic

        :raises: :class:`whad.ble.exceptions.InvalidHandleValueException`
        """
        try:
            if handle in self.__service_by_characteristic_handle:
                return self.__service_by_characteristic_handle[handle]
            else:
                raise InvalidHandleValueException
        except IndexError:
            raise InvalidHandleValueException


    def services(self) -> Iterator[Service]:
        """Enumerate service objects.

        This method is a generator and will yield service objects registered
        into the profile.
        """
        for handle in self.__attr_db:
            object = self.__attr_db[handle]
            if isinstance(object, BlePrimaryService) or isinstance(object, BleSecondaryService):
                yield object

    def included_services(self) -> Iterator[BleIncludeService]:
        for handle in self.__attr_db:
            object = self.__attr_db[handle]
            if isinstance(object, BleIncludeService):
                yield object

    def get_service_by_UUID(self, service_uuid: UUID):
        """Get a service by its UUID.

        :param      service_uuid:   Service UUID to look for
        :type       service_uuid:   :class:`whad.ble.profile.attribute.UUID`
        :return:    Service if found, ``None`` otherwise
        :rtype:     :class:`whad.ble.profile.service.Service`
        """
        for handle in self.__attr_db:
            object = self.__attr_db[handle]
            if isinstance(object, BlePrimaryService) or isinstance(object, BleSecondaryService):
                if object.uuid == service_uuid:
                    return object

        # Not found
        return None

    def get_characteristic_by_UUID(self, charac_uuid: UUID):
        """Get characteristic by its UUID.

        :param      charac_uuid:   Characteristic UUID to look for
        :type       charac_uuid:   :class:`whad.ble.profile.attribute.UUID`
        :return:    Characteristic if found, ``None`` otherwise
        :rtype:     :class:`whad.ble.profile.characteristic.Characteristic`
        """
        for handle in self.__attr_db:
            object = self.__attr_db[handle]
            if isinstance(object, BleCharacteristic):
                if object.uuid == charac_uuid:
                    return object


    def attr_by_type_uuid(self, uuid, start=1, end=0xFFFF) -> Iterator[Attribute]:
        """Enumerate attributes that have a specific type UUID.

        :param  uuid:   Type UUID
        :type   uuid:   :class:`whad.ble.profile.attribute.UUID`
        :param  start:  Start handle
        :type   start:  int
        :param  end:    End handle
        :type   end:    int
        """
        for handle in self.__attr_db:
            object = self.__attr_db[handle]
            if object.type_uuid == uuid and object.handle >= start and object.handle <= end:
                yield object

    def export_json(self):
        """Export profile as JSON data, including services, characteristics and descriptors
        definition.

        :return:    JSON data corresponding to this profile
        :rtype:     str
        """
        profile_dict = {}
        profile_dict['services'] = []
        for service in self.services():
            service_dict = {
                'uuid': str(service.uuid),
                'type_uuid': str(service.type_uuid),
                'start_handle': service.handle,
                'end_handle': service.end_handle
            }
            service_dict['characteristics'] = []
            for charac in service.characteristics():
                charac_dict = {
                    'handle': charac.handle,
                    'uuid': str(charac.type_uuid),
                    'properties': charac.properties,
                    'security': SecurityAccess.accesses_to_int(charac.security),
                    'value': {
                        'handle': charac.value_handle,
                        'uuid': str(charac.uuid),
                    }
                }
                charac_dict['descriptors'] = []
                for desc in charac.descriptors():
                    desc_dict = {
                        'handle': desc.handle,
                        'uuid': str(desc.type_uuid)
                    }
                    charac_dict['descriptors'].append(desc_dict)
                service_dict['characteristics'].append(charac_dict)
            profile_dict['services'].append(service_dict)
        return json.dumps(profile_dict)


    def find_hook(self, service, characteristic, operation) -> callable:
        """Find a registered hook for a specific service, characteristic and operation.

        :param  service:        Service object
        :type   service:        :class:`whad.ble.profile.service.Service`
        :param  characteristic: Characteristic object
        :type   characteristic: :class:`whad.ble.profile.characteristic.Characteristic`
        :param  operation:      GATT operation
        :type   operation:      str

        :return: Hook callback
        :rtype: callable
        """
        hook_key = str(service.uuid) + ':' + str(characteristic.uuid)
        if hook_key in self.__hooks:
            if operation in self.__hooks[hook_key]:
                return self.__hooks[hook_key][operation]
        return None


    ################################################
    # Connection/disconnection hooks
    ################################################

    def on_connect(self, conn_handle):
        """Connection hook.

        This hook is only used to notify the connection of a device.

        :param  conn_handle:    Connection handle
        :type   conn_handle:    int
        """
        pass

    def on_disconnect(self, conn_handle):
        """Disconnection hook.

        This hook is only used to notify the disconnection of a device.

        :param  conn_handle:    Connection handle
        :type   conn_handle:    int
        """
        pass


    ################################################
    # Characteristic Read/Write/Subscribe hooks
    ################################################

    def on_characteristic_read(self, service, characteristic, offset=0, length=0):
        """Characteristic read hook.

        This hook is called whenever a characteristic is about to be read by a GATT client.
        If this method returns a byte array, this byte array will be sent back to the
        GATT client. If this method returns None, then the read operation will return an
        error (not allowed to read characteristic value).


        :param  service:        Service owning the characteristic
        :type   service:        :class:`whad.ble.profile.service.Service`
        :param  characteristic: Characteristic object
        :type   characteristic: :class:`whad.ble.profile.characteristic.Characteristic`
        :param  offset:         Read offset (default: 0)
        :type   offset:         int
        :param  length:         Max read length
        :type   length:         int

        :return:    Value to return to the GATT client
        :rtype:     bytes
        """
        # Check if we have a hook to call
        hook = self.find_hook(service, characteristic, 'read')
        if hook is not None:
            return hook(offset, length)

        # If no hook registered, then return the characteristic value
        return characteristic.value[offset:offset + length]

    def on_characteristic_write(self, service, characteristic, offset=0, value=b'', without_response=False):
        """Characteristic write hook

        This hook is called whenever a charactertistic is about to be written by a GATT
        client.

        :param  service:            Service owning the characteristic
        :type   service:            :class:`whad.ble.profile.service.Service`
        :param  characteristic:     Characteristic object
        :type   characteristic:     :class:`whad.ble.profile.characteristic.Characteristic`
        :param  offset:             Read offset (default: 0)
        :type   offset:             int
        :param  value:              Value about to be written into the characteristic
        :type   value:              bytes
        :param  without_response:   Set to ``True`` if no response is required
        :type   without_response:   bool
        """
        hook = self.find_hook(service, characteristic, 'write')
        if hook is not None:
            return hook(
                offset,
                value,
                without_response=without_response
            )

    def on_characteristic_written(self, service, characteristic, offset=0, value=b'', without_response=False):
        """Characteristic written hook

        This hook is called whenever a charactertistic has been written by a GATT
        client.

        :param  service:            Service owning the characteristic
        :type   service:            :class:`whad.ble.profile.service.Service`
        :param  characteristic:     Characteristic object
        :type   characteristic:     :class:`whad.ble.profile.characteristic.Characteristic`
        :param  offset:             Read offset (default: 0)
        :type   offset:             int
        :param  value:              Value about to be written into the characteristic
        :type   value:              bytes
        :param  without_response:   Set to ``True`` if no response is required
        :type   without_response:   bool
        """
        # Check if we have a hook to call
        hook = self.find_hook(service, characteristic, 'written')
        if hook is not None:
            # Call our hook
            return hook(
                offset,
                value,
                without_response=without_response
            )

    def on_characteristic_subscribed(self, service, characteristic, notification=False, indication=False):
        """Characteristic subscribed hook

        This hook is called whenever a characteristic has been subscribed to.

        :param  service:            Service owning the characteristic
        :type   service:            :class:`whad.ble.profile.service.Service`
        :param  characteristic:     Characteristic object
        :type   characteristic:     :class:`whad.ble.profile.characteristic.Characteristic`
        :param  notification:       Set to ``True`` if subscribed to notification
        :type   notification:       bool
        :param  indication:         Set to ``True`` if subscribed to notification
        :type   indication:         bool
        """
        # Check if we have a hook to call
        hook = self.find_hook(service, characteristic, 'sub')
        if hook is not None:
            # Call our hook
            return hook(
                notification=notification,
                indication=indication
            )

    def on_characteristic_unsubscribed(self, service, characteristic):
        """Characteristic unsubscribed hook

        This hook is called whenever a characteristic has been unsubscribed.

        :param  service:            Service owning the characteristic
        :type   service:            :class:`whad.ble.profile.service.Service`
        :param  characteristic:     Characteristic object
        :type   characteristic:     :class:`whad.ble.profile.characteristic.Characteristic`
        """
        # Check if we have a hook to call
        hook = self.find_hook(service, characteristic, 'unsub')
        if hook is not None:
            # Call our hook
            return hook()

    def on_notification(self, service, characteristic, value):
        """Characteristic notification hook.

        This hook is called when a notification is sent to a characteristic.

        :param  service:            Service owning the characteristic
        :type   service:            :class:`whad.ble.profile.service.Service`
        :param  characteristic:     Characteristic object
        :type   characteristic:     :class:`whad.ble.profile.characteristic.Characteristic`
        :param  value:              Characteristic value
        :type   value:              bytes
        """
        pass

    def on_indication(self, service, characteristic, value):
        """Characteristic indication hook.

        This hook is called when a indication is sent to a characteristic.

        :param  service:            Service owning the characteristic
        :type   service:            :class:`whad.ble.profile.service.Service`
        :param  characteristic:     Characteristic object
        :type   characteristic:     :class:`whad.ble.profile.characteristic.Characteristic`
        :param  value:              Characteristic value
        :type   value:              bytes
        """
        pass



'''
This module provides some helpers functions and constants related to Bluetooth Low Energy physical layer.
'''
from whad.helpers import swap_bits
from whad.phy import Endianness, GFSKModulationScheme, PhysicalLayer
from scapy.layers.bluetooth4LE import BTLE
from enum import IntEnum
from dataclasses import dataclass
from struct import pack

class FieldsSize(IntEnum):
    '''
    Size of major BLE fields (in bytes).
    '''
    ACCESS_ADDRESS_SIZE = 4
    HEADER_SIZE = 2
    CRC_SIZE = 3


def frequency_to_channel(frequency):
    '''
    Converts a frequency (in Hz) to the corresponding BLE channel.
    '''
    if not isinstance(frequency,int) or frequency < 2402000000 or frequency > 2480000000:
        return None

    frequency = int(frequency / 1000000)

    freq_offset = frequency - 2400
    if freq_offset == 2:
        channel = 37
    elif freq_offset == 26:
        channel = 38
    elif freq_offset == 80:
        channel = 39
    elif freq_offset <= 24:
        channel = int((freq_offset / 2) - 2)
    else:
        channel = int((freq_offset / 2) - 3)

    return channel

def channel_to_frequency(channel):
    '''
    Converts a BLE channel to the corresponding frequency (in Hz).
    '''
    if not isinstance(channel,int) or channel < 0 or channel > 39:
        return None
    if channel == 37:
        freq_offset = 2
    elif channel == 38:
        freq_offset = 26
    elif channel == 39:
        freq_offset = 80
    elif channel < 11:
        freq_offset = 2 * (channel + 2)
    else:
        freq_offset = 2 * (channel + 3)

    return (2400 + freq_offset) * 1000000

def dewhitening(data, channel):
  '''
  Dewhiten data based on BLE channel.
  '''
  ret = []
  lfsr = swap_bits(channel) | 2

  for d in data:
    d = swap_bits(d)
    for i in 128, 64, 32, 16, 8, 4, 2, 1:
      if lfsr & 0x80:
        lfsr ^= 0x11
        d ^= i

      lfsr <<= 1
      i >>=1
    ret.append(swap_bits(d))

  return bytes(ret)

def whitening(data, channel):
    '''
    Whiten data based on a BLE channel.
    '''
    return dewhitening(data, channel)

def crc(data, init=0x555555):
    '''
    Computes the 24-bit CRC of provided data.
    '''
    ret = [(init >> 16) & 0xff, (init >> 8) & 0xff, init & 0xff]
    for d in data:
        for v in range(8):
            t = (ret[0] >> 7) & 1

            ret[0] <<= 1
            if ret[1] & 0x80:
                ret[0] |= 1

            ret[1] <<= 1
            if ret[2] & 0x80:
                ret[1] |= 1

            ret[2] <<= 1
            if d & 1 != t:
                ret[2] ^= 0x5b
                ret[1] ^= 0x06

            d >>= 1

    ret[0] = swap_bits(ret[0] & 0xFF)
    ret[1] = swap_bits(ret[1] & 0xFF)
    ret[2] = swap_bits(ret[2] & 0xFF)
    return bytes(ret)



def is_access_address_valid(aa):
    '''
    This function checks if the provided access address is valid.
    '''
    a = (aa & 0xff000000)>>24
    b = (aa & 0x00ff0000)>>16
    c = (aa & 0x0000ff00)>>8
    d = (aa & 0x000000ff)
    if a==b and b==c and c==d:
        return False
    if (aa == 0x8E89BED6):
        return True
    bb = aa
    for i in range(0,26):
        if (bb & 0x3F) == 0 or (bb & 0x3F) == 0x3F:
            return False
        bb >>= 1
    bb = aa
    t = 0
    a = (bb & 0x80000000)>>31
    for i in range(30,0,-1):
        if (bb & (1<<i)) >> i != a:
            a = (bb & (1<<i))>>i
            t += 1
            if t>24:
                return False
        if (i<26) and (t<2):
            return False
    return True


@dataclass
class BLEPhysicalLayerConfiguration:
    address : int = 0x8e89bed6
    channel : int = 37
    crc_init : int = 0x555555


def check_crc(pkt, configuration):
    return crc(pkt[FieldsSize.ACCESS_ADDRESS_SIZE:-FieldsSize.CRC_SIZE], configuration.crc_init) == pkt[-FieldsSize.CRC_SIZE:]

def decoding(pkt, configuration):
    # Remove preamble
    while pkt.startswith(b"\xAA"):
        pkt = pkt[1:]

    dewhitened = dewhitening(pkt[FieldsSize.ACCESS_ADDRESS_SIZE:], configuration.channel)
    size = dewhitened[1]
    packet = pkt[:FieldsSize.ACCESS_ADDRESS_SIZE] + dewhitened[:size+ FieldsSize.CRC_SIZE + FieldsSize.HEADER_SIZE]
    return packet


PHYS = {
    "LE-1M": PhysicalLayer(
                modulation=GFSKModulationScheme(deviation=250000),
                datarate=1000000,
                endianness=Endianness.LITTLE,
                frequency_range=(2402000000, 2480000000),
                maximum_packet_size=31,
                synchronization_word=b"\xAA",
                configuration=BLEPhysicalLayerConfiguration(),
                scapy_layer=BTLE,
                frequency_to_channel_function=frequency_to_channel,
                channel_to_frequency_function=channel_to_frequency,
                format_address_function=lambda access_address:pack(">I", access_address),
                integrity_function=check_crc,
                encoding_function=None,
                decoding_function=decoding,
    ),

    "LE-2M": PhysicalLayer(
                modulation=GFSKModulationScheme(deviation=500000),
                datarate=2000000,
                endianness=Endianness.LITTLE,
                frequency_range=(2402000000, 2480000000),
                maximum_packet_size=250,
                synchronization_word=b"\xAA\xAA",
                configuration=BLEPhysicalLayerConfiguration(),
                scapy_layer=BTLE,
                frequency_to_channel_function=frequency_to_channel,
                channel_to_frequency_function=channel_to_frequency,
                format_address_function=lambda access_address:pack(">I", access_address),
                integrity_function=crc,
                encoding_function=whitening,
                decoding_function=dewhitening
            )
}



"""This module provides a UartDevice class that can be used with a WhadDeviceConnector
to interact with a WHAD-enable firmware device that uses UART as its transport layer.

This class handles device connection, disconnection and read/write operations. All the
parsing magic is performed in our WhadDevice class.
"""

from asyncio import QueueEmpty
import os
import select
from threading import Lock
from serial import Serial
from serial.tools.list_ports import comports
from time import sleep
from queue import Empty

from whad.device import WhadDevice
from whad.exceptions import WhadDeviceNotReady, WhadDeviceError
from whad.protocol.whad_pb2 import Message
from whad.helpers import message_filter
from whad.protocol.device_pb2 import DeviceResetQuery
from whad.protocol.generic_pb2 import ResultCode
from whad.exceptions import WhadDeviceNotFound

from whad.hub.generic.cmdresult import CommandResult, Success
from whad.hub.discovery import DeviceReady

def get_port_info(port):
    """Find information about a serial port

    :param string port: Target serial port
    """
    for p in comports():
        if p.device == port:
            return p
    return None

class UartDevice(WhadDevice):
    """
    UartDevice device class.
    """
    INTERFACE_NAME = "uart"

    @classmethod
    def list(cls):
        '''
        Returns a list of available UART devices.

        To prevent identifying serial ports which are not compatible with WHAD, it implements
        a filtering mechanism based on vid, pid, manufacturer and / or product.
        '''
        supported_uart_devices = (
            (0xc0ff, 0xeeee, "WHAD", "ButteRFly dongle"), # Butterfly Dongle
            (0x303A, None, None, None),   # Espressif ESP-32 board
            (0x10C4, 0xEA60, None, None), # Espressif ESP-32 CP2102 board
            (0x0483, 0x374e, None, None), # Nucleo WL55
        )
        devices = []
        for uart_dev in comports():
            pid, vid, manufacturer, product = uart_dev.pid, uart_dev.vid,uart_dev.manufacturer, uart_dev.product
            for (supported_vid, supported_pid, supported_manufacturer, supported_product) in supported_uart_devices:
                if (
                    (supported_pid is None or supported_pid == pid) and
                    (supported_vid is None or supported_vid == vid) and
                    (supported_manufacturer is None or supported_manufacturer == manufacturer) and
                    (supported_product is None or supported_product == product)
                ):
                    dev = UartDevice(uart_dev.device, baudrate=115200)
                    devices.append(dev)
                    break
        return devices


    def __init__(self, port='/dev/ttyUSB0', baudrate=115200):
        """
        Create device connection
        """
        super().__init__()

        # Connect to target UART device in non-blocking mode
        self.__port = port
        self.__baudrate = baudrate
        self.__fileno = None
        self.__uart = None
        self.__opened = False

        # Determine if device is CDC ACM (usb subsystem)
        port_info = get_port_info(self.__port)
        if port_info is None:
            raise WhadDeviceNotFound
        else:
            self.__is_acm = (port_info.subsystem == 'usb')

    @property
    def identifier(self):
        '''
        Returns the identifier of the device (e.g., serial port).
        '''
        return self.__port

    def is_acm(self):
        """Determine if this UART device is a CDC ACM one.

        :return: True if CDC ACM, False otherwise.
        """
        return self.__is_acm


    def open(self):
        """
        Open device.
        """
        if not self.__opened:
            # Open UART device
            self.__uart = Serial(self.__port, self.__baudrate)

            # Get file number to use with select()
            self.__fileno = self.__uart.fileno()
            self.__opened = True

            # If device is CDC ACM, we don't need to reset it
            # only reset true serial devices
            if not self.__is_acm:
                self.__uart.dtr = False             # Non reset state
                self.__uart.rts = False             # Non reset state
                self.__uart.dtr = self.__uart.dtr   # usbser.sys workaround

            # Ask parent class to run a background I/O thread
            super().open()


    def reset(self):
        """Reset device.

        This routine tries to reset device by setting RTS to high. This works
        on NodeMCU chips as well as any Arduino-compatible device.
        """
        # If device is a true serial device, ask for a reset through DTR/RTS
        if not self.__is_acm:
            # Reset device through DTR
            self.__uart.dtr = False             # Non reset state
            self.__uart.rts = True             # Non reset state
            sleep(0.2)
            self.__uart.dtr = False             # Non reset state
            self.__uart.rts = False             # Non reset state

            try:
                # Wait for a ready message for 1 second
                msg = self.wait_for_single_message(
                    1.0,
                    message_filter(DeviceReady)
                )
                self.dispatch_message(msg)
            except Empty:
                # Use the classic way to reset device (RTS-based reset failed)
                msg = self.hub.discovery.create_reset_query()
                self.send_command(
                    msg,
                    message_filter(DeviceReady)
                )
        else:
            # Device is ACM, send a classic reset message to device
            msg = self.hub.discovery.create_reset_query()
            self.send_command(
                msg,
                message_filter(DeviceReady)
            )


    def close(self):
        """
        Close current device.
        """
        # Ask parent class to stop I/O thread
        super().close()

        # Close underlying device.
        if self.__uart is not None:
            self.__uart.close()
        self.__uart = None
        self.__fileno = None
        self.__opened = False

    def write(self, data):
        """Writes data to the device. It relies on select() in order to make sure
        we are allowed to write to the device and wait without eating too much CPU
        if the device is not ready to be written to.

        :param bytes data: Data to write
        :returns: number of bytes written to the device
        """
        try:
            if not self.__opened:
                raise WhadDeviceNotReady()

            nb_bytes_written = 0
            wlist = [self.__fileno]
            elist = [self.__fileno]
            readers,writers,errors = select.select(
                [],
                wlist,
                elist
            )

            if len(writers) > 0:
                nb_bytes_written = os.write(self.__fileno, data)
            return nb_bytes_written
        except OSError as os_error:
            raise WhadDeviceError("Sending data to WHAD device failed.")

    def read(self):
        """Fetches data from the device, if there is any data to read. We call select()
        to make sure data is waiting to be read before reading it. Data is then sent to
        our parsing method through on_data_received() that will handle data reassembling
        and message parsing and dispatch.
        """

        if not self.__opened:
            raise WhadDeviceNotReady()

        rlist = [self.__fileno]
        wlist = []
        elist = [self.__fileno]

        readers,writers,errors = select.select(
            rlist,
            wlist,
            elist,
            1
        )

        if not self.__opened:
            raise WhadDeviceNotReady()

        # Handle incoming messages if any
        if len(readers) > 0 and self.__fileno is not None:
            data = os.read(self.__fileno, 1024)
            self.on_data_received(data)

    def change_transport_speed(self, speed):
        """Set UART speed for true serial devices. CDC ACM devices will ignore
        it.

        :param int speed: New baudrate to apply to current serial device.
        """
        if not self.__is_acm:
            #msg = Message()
            #msg.discovery.set_speed.speed = speed
            msg = self.hub.discovery.create_set_speed(speed)
            resp = self.send_command(
                msg,
                message_filter(CommandResult)
            )

            if isinstance(resp, Success):
                # Change baudrate
                self.__uart.baudrate = speed

                # Wait for device to be ready (500ms)
                sleep(0.5)



"""
Bluetooth Low Energy Peripheral connector
=========================================

WHAD provides a specific connector to create a BLE device, :class:`Peripheral`.
This connector implements a GATT server and hosts a GATT profile, allowing remote
BLE devices to connect to it and query its services, characteristics, and descriptors.

The connector provides some callbacks such as :meth:`Peripheral.on_connected` to
react on specific events.
"""

from time import sleep

from whad.ble.connector import BLE
from whad.hub.ble.bdaddr import BDAddress
from whad.ble.stack import BleStack
from whad.ble.stack.gatt import GattServer, GattClientServer
from whad.ble.stack.att import ATTLayer
from whad.ble.stack.smp import CryptographicDatabase, Pairing
from whad.ble.profile import GenericProfile
from whad.ble.profile.device import PeripheralDevice
from whad.ble.profile.advdata import AdvDataFieldList, AdvFlagsField
from whad.hub.ble import Direction as BleDirection
from whad.exceptions import UnsupportedCapability
from time import sleep

from binascii import hexlify

# Logging
import logging
logger = logging.getLogger(__name__)

class Peripheral(BLE):
    """This BLE connector provides a way to create a peripheral device.

    A peripheral device exposes some services and characteristics that may
    be accessed by a central device. These services and characteristics are
    defined by a specific profile.
    """

    def __init__(self, device, existing_connection = None, profile=None, adv_data=None, scan_data=None, bd_address=None, public=True, stack=BleStack, gatt=GattServer, pairing=Pairing(), security_database=None):
        """Create a peripheral device.

        :param  device:     WHAD device to use as a peripheral
        :type   device:     :class:`whad.device.WhadDevice`
        :param  profile:    Device profile to use
        :type   profile:    :class:`whad.ble.profile.GenericProfile`
        :param  adv_data:   Advertisement data
        :type   adv_data:   :class:`whad.ble.profile.advdata.AdvDataFieldList`
        :param  scan_data:  Advertisement data sent in Scan Response
        :type   scan_data:  :class:`whad.ble.profile.advdata.AdvDataFieldList`
        :param  bd_address: Bluetooth Device address to use
        :type   bd_address: str
        :param  public:     Set to True to use a public Bluetooth Device address, False to use a random one
        :type   public:     bool
        :param  stack:      Bluetooth Low Energy stack to use, :class:`whad.ble.stack.BleStack` by default
        :param  gatt:       Bluetooth Low Energy GATT
        """
        super().__init__(device)

        # Attach a GATT server to our stack ATT layer
        att_layer = stack.find('att')
        if att_layer is not None:
            att_layer.add(gatt)

        # Initialize local peer and remote per info
        self.__local_peer = None
        self.__remote_peer = None

        # Initialize stack
        self.__stack = stack(self)
        self.connection = None
        self.__connected = False
        self.__conn_handle = None

        # Initialize profile
        if profile is None:
            logger.info('No profile provided to this Peripheral instance, use a default one.')
            self.__profile = GenericProfile()
        else:
            logger.info('Peripheral will use the provided profile.')
            self.__profile = profile

        # Initialize security database
        if security_database is None:
            logger.info('No security database provided to this Peripheral instance, use a default one.')
            self.__security_database = CryptographicDatabase()
        else:
            logger.info('Peripheral will use the provided security database.')
            self.__security_database = security_database

        # Initiate pairing parameters
        self.__pairing_parameters = pairing

        # Check if device accepts peripheral mode
        if not self.can_be_peripheral():
            logger.info('Capability MasterRole not supported by this WHAD device')
            raise UnsupportedCapability("Peripheral")
        else:
            # Set bd address if provided
            if bd_address is not None:
                logger.info('Set BD address to %s' % bd_address)
                self.set_bd_address(bd_address, public=public)

            # If an existing connection is hijacked, simulate a connection
            if existing_connection is not None:
                self.on_connected(existing_connection)
            else:
                # If no advertising data has been set, initialize this peripheral
                # with default flags.
                if adv_data is None:
                    adv_data = AdvDataFieldList(AdvFlagsField())
                
                # Enable peripheral mode
                logger.info('Enable peripheral mode with advertising data: %s' % adv_data)
                self.enable_peripheral_mode(adv_data, scan_data)

    @property
    def local_peer(self):
        return self.__local_peer

    @property
    def remote_peer(self):
        return self.__remote_peer

    @property
    def conn_handle(self):
        return self.__conn_handle

    def get_pairing_parameters(self):
        """Returns the provided pairing parameters, if any.
        """
        return self.__pairing_parameters

    def send_data_pdu(self, pdu, conn_handle=1, direction=BleDirection.SLAVE_TO_MASTER, access_address=0x8e89bed6, encrypt=None) -> bool:
        """Send a PDU to the central device this peripheral device is connected to.

        Sending direction is set to ̀ BleDirection.SLAVE_TO_MASTER` as we need to send PDUs to a central device.

        :param  pdu:            PDU to send
        :type   pdu:            :class:`scapy.layers.bluetooth4LE.BTLE`
        :param  conn_handle:    Connection handle
        :type   conn_handle:    int
        :param  direction:      Sending direction (to master or slave)
        :type   direction:      :class:`whad.protocol.ble_pb2.BleDirection`, optional
        :param  access_address: Target access address
        :type   access_address: int, optional
        :return:                PDU transmission result.
        :rtype: bool
        """
        return super().send_data_pdu(pdu, conn_handle=conn_handle, direction=direction, access_address=access_address, encrypt=encrypt)


    def send_ctrl_pdu(self, pdu, conn_handle=1, direction=BleDirection.SLAVE_TO_MASTER, access_address=0x8e89bed6, encrypt=None) -> bool:
        """Send a PDU to the central device this peripheral device is connected to.

        Sending direction is set to ̀ BleDirection.SLAVE_TO_MASTER` as we need to send PDUs to a central device.

        :param  pdu:            PDU to send
        :type   pdu:            :class:`scapy.layers.bluetooth4LE.BTLE`
        :param  conn_handle:    Connection handle
        :type   conn_handle:    int
        :param  direction:      Sending direction (to master or slave)
        :type   direction:      :class:`whad.protocol.ble_pb2.BleDirection`, optional
        :param  access_address: Target access address
        :type   access_address: int, optional
        :return:                PDU transmission result.
        :rtype: bool
        """
        return super().send_ctrl_pdu(pdu, conn_handle=conn_handle, direction=direction, access_address=access_address, encrypt=encrypt)


    def use_stack(self, clazz=BleStack):
        """Specify a stack class to use for BLE. By default, our own stack (BleStack) is used.

        :param  clazz:  BLE stack to use.
        :type   clazz:  :class:`whad.ble.stack.BleStack`
        """
        self.__stack = clazz(self)

    @property
    def smp(self):
        if self.connection is not None:
            return self.connection.smp
        return None

    def pairing(self, pairing=None):
        if self.smp is None:
            return False
        if pairing is not None:
            self.__pairing_parameters = pairing

        if not self.smp.request_pairing(self.__pairing_parameters):
            return False

        while not self.smp.is_pairing_done():
            sleep(0.1)
            if self.smp.is_pairing_failed():
                return False

        self.smp.reset_state()
        return True

    @property
    def gatt(self):
        if self.connection is not None:
            return self.connection.gatt
        return None

    @property
    def security_database(self):
        return self.__security_database

    def is_connected(self) -> bool:
        """Determine if the peripheral has an active connection from a
        GATT client.
        """
        return self.__connected

    def wait_connection(self):
        """Wait for a GATT client to connect to the peripheral. If a connection
        is already active, returns immediately.
        """
        while not self.is_connected():
            sleep(.5)


    ##############################
    # Incoming events
    ##############################

    def on_connected(self, connection_data):
        """A device has just connected to this peripheral.

        :param  connection_data:    Connection data
        :type   connection_data:    :class:`whad.protocol.ble_pb2.Connected`
        """
        # Retrieve the GATT server instance and set its profile
        logger.info('a device is now connected (connection handle: %d)' % connection_data.conn_handle)
        self.__local_peer = BDAddress.from_bytes(
            connection_data.advertiser,
            connection_data.adv_addr_type
        )
        self.__remote_peer = BDAddress.from_bytes(
            connection_data.initiator,
            connection_data.init_addr_type
        )
        self.__stack.on_connection(
            connection_data.conn_handle,
            self.__local_peer,
            self.__remote_peer
        )

        # GATT server is now connected
        self.__connected = True
        self.__conn_handle = connection_data.conn_handle

    def on_disconnected(self, disconnection_data):
        """A device has just disconnected from this peripheral.

        :param  connection_data:    Connection data
        :type   connection_data:    :class:`whad.protocol.ble_pb2.Disconnected`
        """

        logger.info('a device has just connected (connection handle: %d)' % disconnection_data.conn_handle)
        self.__stack.on_disconnection(
            disconnection_data.conn_handle,
            disconnection_data.reason
        )

        # Notify peripheral device about this disconnection
        if self.__profile is not None:
            self.__profile.on_disconnect(disconnection_data.conn_handle)

        # We are now disconnected
        self.__connected = False


    def on_ctl_pdu(self, pdu):
        """This method is called whenever a control PDU is received.
        This PDU is then forwarded to the BLE stack to handle it.

        Peripheral devices act as a slave, so we only forward master to slave
        messages to the stack.

        :param  pdu:    BLE PDU
        :type   pdu:    :class:`scapy.layers.bluetooth4LE.BTLE`
        """
        if pdu.metadata.direction == BleDirection.MASTER_TO_SLAVE:
            logger.info('Control PDU comes from master, forward to peripheral')
            self.__stack.on_ctl_pdu(pdu.metadata.connection_handle, pdu)

    def on_data_pdu(self, pdu):
        """This method is called whenever a data PDU is received.
        This PDU is then forwarded to the BLE stack to handle it.

        :param  pdu:    BLE PDU
        :type   pdu:    :class:`scapy.layers.bluetooth4LE.BTLE_DATA`
        """
        if pdu.metadata.direction == BleDirection.MASTER_TO_SLAVE:
            logger.info('Data PDU comes from master, forward to peripheral')
            self.__stack.on_data_pdu(pdu.metadata.connection_handle, pdu)

    def on_new_connection(self, connection):
        """On new connection, discover primary services

        :param  connection:    Connection data
        :type   connection:    :class:`whad.protocol.ble_pb2.Connected`
        """
        # Use GATT server
        self.connection = connection
        self.__connected = True

        # Retrieve GATT server
        self.__gatt_server = connection.gatt

        self.__gatt_server.set_server_model(self.__profile)
        self.__connected = True

        # Configure SMP layer
        # we set the security database
        self.connection.smp.set_security_database(self.__security_database)
        self.connection.smp.pairing_parameters = self.__pairing_parameters

        # Check if we got a matching LTK
        crypto_material = self.security_database.get(address=self.__local_peer)

        if crypto_material is not None and crypto_material.has_ltk():
            conn_handle = connection.conn_handle
            self.__stack.get_layer('ll').state.register_encryption_key(
                conn_handle,
                crypto_material.ltk.value
            )
            if crypto_material.is_authenticated():
                #print("Marked as authenticated")
                self.__stack.get_layer('ll').state.mark_as_authenticated(connection.conn_handle)
            else:
                pass#print("Marked as unauthenticated")


        # we indicate that we are a responder
        self.connection.smp.set_responder_role()

        # Notify our profile about this connection
        self.__profile.on_connect(self.connection.conn_handle)
    '''
    def start_encryption(self):
        # Check if we got a matching LTK
        crypto_material = self.security_database.get(address=self.__target)

        if crypto_material is not None and crypto_material.has_ltk():
            conn_handle = self.connection.smp.get_layer('l2cap').state.conn_handle
            if crypto_material is not None and crypto_material.has_ltk():
                self.connection.smp.get_layer('ll').start_encryption(
                    conn_handle,
                    unpack('>Q', crypto_material.ltk.rand)[0],
                    crypto_material.ltk.ediv
                )
    '''
class PeripheralClient(Peripheral):
    '''This BLE connector provides a way to create a peripheral device with
    both GATT server and client roles.
    '''

    def __init__(self, device, existing_connection = None, profile=None, adv_data=None, scan_data=None, bd_address=None, public=True, stack=BleStack):
        super().__init__(
            device,
            existing_connection=existing_connection,
            profile=profile,
            adv_data=adv_data,
            scan_data=scan_data,
            bd_address=bd_address,
            public=public,
            stack=stack
        )

        # Change ATTLayer to use GattClientServer and reinstantiate our stack
        ATTLayer.add(GattClientServer)
        self.use_stack(BleStack)


    def on_new_connection(self, connection):
        super().on_new_connection(connection)

        # Create a new peripheral device to represent the central device
        # that has just connected

        self.__peripheral = PeripheralDevice(
            self,
            connection.gatt,
            connection.conn_handle
        )

        # Retrieve GATT client
        self.__central = connection.gatt
        self.__central.set_client_model(self.__peripheral)

    @property
    def central_device(self):
        return self.__peripheral
